#include "Python.h"
#include "numpy/arrayobject.h"
#include "subgridError.h"
/** \file csubgridErrorModule.c
    \defgroup csubgridError csubgridError
    \brief Python interface to subgridError library
    @{
*/
#define ND(p) ((PyArrayObject *)p)->nd
#define SHAPE(p) ((PyArrayObject *)p)->dimensions
#define DDATA(p) ((double *) (((PyArrayObject *)p)->data))
#define IDATA(p) ((int *) (((PyArrayObject *)p)->data))

static PyObject*
csubgridErrorCalculateSubgridError_tauRes(PyObject* self,
                                          PyObject* args)
{
  PyObject *tau,*pdeResidual,*dpdeResidual,*subgridError,*dsubgridError;
  if(!PyArg_ParseTuple(args,"OOOOO",
                       &tau,
                       &pdeResidual,
                       &dpdeResidual,
                       &subgridError,
                       &dsubgridError))
    return NULL;         
  calculateSubgridError_tauRes(SHAPE(dsubgridError)[0],
                               SHAPE(dsubgridError)[1],
                               SHAPE(dsubgridError)[2],
                               DDATA(tau),
                               DDATA(pdeResidual),
                               DDATA(dpdeResidual),
                               DDATA(subgridError),
                               DDATA(dsubgridError));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorGradient_tauRes(PyObject* self,
						  PyObject* args)
{
  PyObject *tau_gradient,*grad_pdeResidual,*grad_subgridError;
  if(!PyArg_ParseTuple(args,"OOO",
                       &tau_gradient,
                       &grad_pdeResidual,
                       &grad_subgridError))
    return NULL;         
  calculateSubgridErrorGradient_tauRes(SHAPE(grad_subgridError)[0],
				       SHAPE(grad_subgridError)[1],
				       SHAPE(grad_subgridError)[2],
				       DDATA(tau_gradient),
				       DDATA(grad_pdeResidual),
				       DDATA(grad_subgridError));

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridError_ADR_tau(PyObject* self,
                                                 PyObject* args)
{
  PyObject *elementDiameter,*dmt,*df,*a,*da,*grad_phi,*dphi,*dr,
    *pe,*cfl,*tau;
  char stabilizationType;
  if(!PyArg_ParseTuple(args,"cOOOOOOOOOOO",
                       &stabilizationType,
                       &elementDiameter,
                       &dmt,
                       &df,
                       &a,
                       &da,
                       &grad_phi,
                       &dphi,
                       &dr,
                       &pe,
                       &cfl,
                       &tau))
    return NULL;
  calculateSubgridError_ADR_tau(SHAPE(df)[0],
                                SHAPE(df)[1],
                                SHAPE(df)[2],
                                stabilizationType,
                                DDATA(elementDiameter),
                                DDATA(dmt),
                                DDATA(df),
                                DDATA(a),
                                DDATA(da),
                                DDATA(grad_phi),
                                DDATA(dphi),
                                DDATA(dr),
                                DDATA(pe),
                                DDATA(cfl),
                                DDATA(tau));
  Py_INCREF(Py_None);
  return Py_None;
}
static PyObject*
csubgridErrorCalculateSubgridError_ADR_tau_sd(PyObject* self,
					      PyObject* args)
{
  PyObject *elementDiameter,*dmt,*df,*a,*da,*grad_phi,*dphi,*dr,
    *pe,*cfl,*tau,*rowptr,*colind;
  char stabilizationType;
  if(!PyArg_ParseTuple(args,"cOOOOOOOOOOOOO",
                       &stabilizationType,
		       &rowptr,
		       &colind,
                       &elementDiameter,
                       &dmt,
                       &df,
                       &a,
                       &da,
                       &grad_phi,
                       &dphi,
                       &dr,
                       &pe,
                       &cfl,
                       &tau))
    return NULL;
  calculateSubgridError_ADR_tau_sd(SHAPE(df)[0],
				   SHAPE(df)[1],
				   SHAPE(df)[2],
				   stabilizationType,
				   IDATA(rowptr),
				   IDATA(colind),
				   DDATA(elementDiameter),
				   DDATA(dmt),
				   DDATA(df),
				   DDATA(a),
				   DDATA(da),
				   DDATA(grad_phi),
				   DDATA(dphi),
				   DDATA(dr),
				   DDATA(pe),
				   DDATA(cfl),
				   DDATA(tau));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridError_ADR_generic_tau(PyObject* self,
                                                   PyObject* args)
{
  PyObject *inverseJ,*dmt,*df,*a,*da,*grad_phi,*dphi,*dr,
    *pe,*cfl,*tau;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOO",
                       &inverseJ,
                       &dmt,
                       &df,
                       &a,
                       &da,
                       &grad_phi,
                       &dphi,
                       &dr,
                       &pe,
                       &cfl,
                       &tau))
    return NULL;
  calculateSubgridError_ADR_generic_tau(SHAPE(df)[0],
                                        SHAPE(df)[1],
                                        SHAPE(df)[2],
                                        DDATA(inverseJ),
                                        DDATA(dmt),
                                        DDATA(df),
                                        DDATA(a),
                                        DDATA(da),
                                        DDATA(grad_phi),
                                        DDATA(dphi),
                                        DDATA(dr),
                                        DDATA(pe),
                                        DDATA(cfl),
                                        DDATA(tau));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridError_ADR_generic_tau_sd(PyObject* self,
                                                   PyObject* args)
{
  PyObject *inverseJ,*dmt,*df,*a,*da,*grad_phi,*dphi,*dr,
    *pe,*cfl,*tau,*rowptr,*colind;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOO",
		       &rowptr,
		       &colind,
                       &inverseJ,
                       &dmt,
                       &df,
                       &a,
                       &da,
                       &grad_phi,
                       &dphi,
                       &dr,
                       &pe,
                       &cfl,
                       &tau))
    return NULL;
  calculateSubgridError_ADR_generic_tau_sd(SHAPE(df)[0],
					   SHAPE(df)[1],
					   SHAPE(df)[2],
					   IDATA(rowptr),
					   IDATA(colind),
					   DDATA(inverseJ),
					   DDATA(dmt),
					   DDATA(df),
					   DDATA(a),
					   DDATA(da),
					   DDATA(grad_phi),
					   DDATA(dphi),
					   DDATA(dr),
					   DDATA(pe),
					   DDATA(cfl),
					   DDATA(tau));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridError_ADR_Sangalli_tau_sd(PyObject* self,
						       PyObject* args)
{
  PyObject *inverseJ,*dmt,*df,*a,*da,*grad_phi,*dphi,*dr,
    *pe,*cfl,*tau,*tau_gradient,*rowptr,*colind;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOO",
		       &rowptr,
		       &colind,
                       &inverseJ,
                       &dmt,
                       &df,
                       &a,
                       &da,
                       &grad_phi,
                       &dphi,
                       &dr,
                       &pe,
                       &cfl,
                       &tau,
		       &tau_gradient))
    return NULL;
  calculateSubgridError_ADR_Sangalli_tau_sd(SHAPE(df)[0],
					    SHAPE(df)[1],
					    SHAPE(df)[2],
					    IDATA(rowptr),
					    IDATA(colind),
					    DDATA(inverseJ),
					    DDATA(dmt),
					    DDATA(df),
					    DDATA(a),
					    DDATA(da),
					    DDATA(grad_phi),
					    DDATA(dphi),
					    DDATA(dr),
					    DDATA(pe),
					    DDATA(cfl),
					    DDATA(tau),
					    DDATA(tau_gradient));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridError_A_tau(PyObject* self,
                                                 PyObject* args)
{
  PyObject *elementDiameter,*dmt,*df,
    *cfl,*tau;
  char stabilizationType;
  if(!PyArg_ParseTuple(args,"cOOOOO",
                       &stabilizationType,
                       &elementDiameter,
                       &dmt,
                       &df,
                       &cfl,
                       &tau))
    return NULL;
  calculateSubgridError_A_tau(SHAPE(df)[0],
                              SHAPE(df)[1],
                              SHAPE(df)[2],
                              stabilizationType,
                              DDATA(elementDiameter),
                              DDATA(dmt),
                              DDATA(df),
                              DDATA(cfl),
                              DDATA(tau));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridError_HJ_tau(PyObject* self,
                                                 PyObject* args)
{
  PyObject *elementDiameter,*dmt,*dH,
    *cfl,*tau;
  char stabilizationType;
  if(!PyArg_ParseTuple(args,"cOOOOO",
                       &stabilizationType,
                       &elementDiameter,
                       &dmt,
                       &dH,
                       &cfl,
                       &tau))
    return NULL;
  /* use same formulat as advection with df = dH */
  calculateSubgridError_A_tau(SHAPE(dH)[0],
                              SHAPE(dH)[1],
                              SHAPE(dH)[2],
                              stabilizationType,
                              DDATA(elementDiameter),
                              DDATA(dmt),
                              DDATA(dH),
                              DDATA(cfl),
                              DDATA(tau));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorStokes2D_1(PyObject* self,
                                             PyObject* args)
{
  PyObject *elementDiameter,*u,*v,*a,
    *pdeResidualP,
    *dpdeResidualP_du,
    *dpdeResidualP_dv,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_dv,
    *subgridErrorP,
    *dsubgridErrorP_dp,
    *dsubgridErrorP_du,
    *dsubgridErrorP_dv,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *dsubgridErrorU_dv,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_du,
    *dsubgridErrorV_dv;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOOOOOOOOOOOOO",
                       &elementDiameter,
                       &u,
                       &v,
                       &a,
                       &pdeResidualP,
                       &dpdeResidualP_du,
                       &dpdeResidualP_dv,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_dv,
                       &subgridErrorP,
                       &dsubgridErrorP_dp,
                       &dsubgridErrorP_du,
                       &dsubgridErrorP_dv,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &dsubgridErrorU_dv,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_du,
                       &dsubgridErrorV_dv))
    return NULL;
  calculateSubgridErrorStokes2D_1(SHAPE(dsubgridErrorP_du)[0],
                                  SHAPE(dsubgridErrorP_du)[1],
                                  SHAPE(dsubgridErrorP_du)[2],
                                  2,
                                  DDATA(elementDiameter),
                                  DDATA(u),
                                  DDATA(v),
                                  DDATA(a),
                                  DDATA(pdeResidualP),
                                  DDATA(dpdeResidualP_du),
                                  DDATA(dpdeResidualP_dv),
                                  DDATA(pdeResidualU),
                                  DDATA(dpdeResidualU_dp),
                                  DDATA(dpdeResidualU_du),
                                  DDATA(pdeResidualV),
                                  DDATA(dpdeResidualV_dp),
                                  DDATA(dpdeResidualV_dv),
                                  DDATA(subgridErrorP),
                                  DDATA(dsubgridErrorP_dp),
                                  DDATA(dsubgridErrorP_du),
                                  DDATA(dsubgridErrorP_dv),
                                  DDATA(subgridErrorU),
                                  DDATA(dsubgridErrorU_dp),
                                  DDATA(dsubgridErrorU_du),
                                  DDATA(dsubgridErrorU_dv),
                                  DDATA(subgridErrorV),
                                  DDATA(dsubgridErrorV_dp),
                                  DDATA(dsubgridErrorV_du),
                                  DDATA(dsubgridErrorV_dv));
   Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorStokes2D_1_sd(PyObject* self,
						PyObject* args)
{
  PyObject *elementDiameter,*u,*v,*a,
    *pdeResidualP,
    *dpdeResidualP_du,
    *dpdeResidualP_dv,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_dv,
    *subgridErrorP,
    *dsubgridErrorP_dp,
    *dsubgridErrorP_du,
    *dsubgridErrorP_dv,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *dsubgridErrorU_dv,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_du,
    *dsubgridErrorV_dv;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOOOOOOOOOOOOO",
                       &elementDiameter,
                       &u,
                       &v,
                       &a,
                       &pdeResidualP,
                       &dpdeResidualP_du,
                       &dpdeResidualP_dv,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_dv,
                       &subgridErrorP,
                       &dsubgridErrorP_dp,
                       &dsubgridErrorP_du,
                       &dsubgridErrorP_dv,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &dsubgridErrorU_dv,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_du,
                       &dsubgridErrorV_dv))
    return NULL;
  calculateSubgridErrorStokes2D_1_sd(SHAPE(dsubgridErrorP_du)[0],
                                  SHAPE(dsubgridErrorP_du)[1],
                                  SHAPE(dsubgridErrorP_du)[2],
                                  2,
                                  DDATA(elementDiameter),
                                  DDATA(u),
                                  DDATA(v),
                                  DDATA(a),
                                  DDATA(pdeResidualP),
                                  DDATA(dpdeResidualP_du),
                                  DDATA(dpdeResidualP_dv),
                                  DDATA(pdeResidualU),
                                  DDATA(dpdeResidualU_dp),
                                  DDATA(dpdeResidualU_du),
                                  DDATA(pdeResidualV),
                                  DDATA(dpdeResidualV_dp),
                                  DDATA(dpdeResidualV_dv),
                                  DDATA(subgridErrorP),
                                  DDATA(dsubgridErrorP_dp),
                                  DDATA(dsubgridErrorP_du),
                                  DDATA(dsubgridErrorP_dv),
                                  DDATA(subgridErrorU),
                                  DDATA(dsubgridErrorU_dp),
                                  DDATA(dsubgridErrorU_du),
                                  DDATA(dsubgridErrorU_dv),
                                  DDATA(subgridErrorV),
                                  DDATA(dsubgridErrorV_dp),
                                  DDATA(dsubgridErrorV_du),
                                  DDATA(dsubgridErrorV_dv));
   Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorStokes2D_GLS_velocity(PyObject* self,
                                                        PyObject* args)
{
  PyObject *elementDiameter,
    *a,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_dv,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_dv;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOO",
                       &elementDiameter,
                       &a,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_dv,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_dv))
    return NULL;
  calculateSubgridErrorStokes2D_GLS_velocity(SHAPE(dsubgridErrorU_du)[0],
                                             SHAPE(dsubgridErrorU_du)[1],
                                             SHAPE(dsubgridErrorU_du)[2],
                                             2,
                                             DDATA(elementDiameter),
                                             DDATA(a),
                                             DDATA(pdeResidualU),
                                             DDATA(dpdeResidualU_dp),
                                             DDATA(dpdeResidualU_du),
                                             DDATA(pdeResidualV),
                                             DDATA(dpdeResidualV_dp),
                                             DDATA(dpdeResidualV_dv),
                                             DDATA(subgridErrorU),
                                             DDATA(dsubgridErrorU_dp),
                                             DDATA(dsubgridErrorU_du),
                                             DDATA(subgridErrorV),
                                             DDATA(dsubgridErrorV_dp),
                                             DDATA(dsubgridErrorV_dv));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorStokes2D_GLS_velocity_sd(PyObject* self,
                                                        PyObject* args)
{
  PyObject *elementDiameter,
    *a,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_dv,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_dv;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOO",
                       &elementDiameter,
                       &a,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_dv,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_dv))
    return NULL;
  calculateSubgridErrorStokes2D_GLS_velocity_sd(SHAPE(dsubgridErrorU_du)[0],
                                             SHAPE(dsubgridErrorU_du)[1],
                                             SHAPE(dsubgridErrorU_du)[2],
                                             2,
                                             DDATA(elementDiameter),
                                             DDATA(a),
                                             DDATA(pdeResidualU),
                                             DDATA(dpdeResidualU_dp),
                                             DDATA(dpdeResidualU_du),
                                             DDATA(pdeResidualV),
                                             DDATA(dpdeResidualV_dp),
                                             DDATA(dpdeResidualV_dv),
                                             DDATA(subgridErrorU),
                                             DDATA(dsubgridErrorU_dp),
                                             DDATA(dsubgridErrorU_du),
                                             DDATA(subgridErrorV),
                                             DDATA(dsubgridErrorV_dp),
                                             DDATA(dsubgridErrorV_dv));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorStokes3D_GLS_velocity(PyObject* self,
                                                        PyObject* args)
{
  PyObject *elementDiameter,
    *a,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_dv,
    *pdeResidualW,
    *dpdeResidualW_dp,
    *dpdeResidualW_dw,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_dv,
    *subgridErrorW,
    *dsubgridErrorW_dp,
    *dsubgridErrorW_dw;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOOOOOOOO",
                       &elementDiameter,
                       &a,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_dv,
                       &pdeResidualW,
                       &dpdeResidualW_dp,
                       &dpdeResidualW_dw,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_dv,
                       &subgridErrorW,
                       &dsubgridErrorW_dp,
                       &dsubgridErrorW_dw))
    return NULL;
  calculateSubgridErrorStokes3D_GLS_velocity(SHAPE(dsubgridErrorU_du)[0],
                                             SHAPE(dsubgridErrorU_du)[1],
                                             SHAPE(dsubgridErrorU_du)[2],
                                             3,
                                             DDATA(elementDiameter),
                                             DDATA(a),
                                             DDATA(pdeResidualU),
                                             DDATA(dpdeResidualU_dp),
                                             DDATA(dpdeResidualU_du),
                                             DDATA(pdeResidualV),
                                             DDATA(dpdeResidualV_dp),
                                             DDATA(dpdeResidualV_dv),
                                             DDATA(pdeResidualW),
                                             DDATA(dpdeResidualW_dp),
                                             DDATA(dpdeResidualW_dw),
                                             DDATA(subgridErrorU),
                                             DDATA(dsubgridErrorU_dp),
                                             DDATA(dsubgridErrorU_du),
                                             DDATA(subgridErrorV),
                                             DDATA(dsubgridErrorV_dp),
                                             DDATA(dsubgridErrorV_dv),
                                             DDATA(subgridErrorW),
                                             DDATA(dsubgridErrorW_dp),
                                             DDATA(dsubgridErrorW_dw));
   Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorStokes3D_GLS_velocity_sd(PyObject* self,
                                                        PyObject* args)
{
  PyObject *elementDiameter,
    *a,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_dv,
    *pdeResidualW,
    *dpdeResidualW_dp,
    *dpdeResidualW_dw,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_dv,
    *subgridErrorW,
    *dsubgridErrorW_dp,
    *dsubgridErrorW_dw;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOOOOOOOO",
                       &elementDiameter,
                       &a,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_dv,
                       &pdeResidualW,
                       &dpdeResidualW_dp,
                       &dpdeResidualW_dw,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_dv,
                       &subgridErrorW,
                       &dsubgridErrorW_dp,
                       &dsubgridErrorW_dw))
    return NULL;
  calculateSubgridErrorStokes3D_GLS_velocity_sd(SHAPE(dsubgridErrorU_du)[0],
                                             SHAPE(dsubgridErrorU_du)[1],
                                             SHAPE(dsubgridErrorU_du)[2],
                                             3,
                                             DDATA(elementDiameter),
                                             DDATA(a),
                                             DDATA(pdeResidualU),
                                             DDATA(dpdeResidualU_dp),
                                             DDATA(dpdeResidualU_du),
                                             DDATA(pdeResidualV),
                                             DDATA(dpdeResidualV_dp),
                                             DDATA(dpdeResidualV_dv),
                                             DDATA(pdeResidualW),
                                             DDATA(dpdeResidualW_dp),
                                             DDATA(dpdeResidualW_dw),
                                             DDATA(subgridErrorU),
                                             DDATA(dsubgridErrorU_dp),
                                             DDATA(dsubgridErrorU_du),
                                             DDATA(subgridErrorV),
                                             DDATA(dsubgridErrorV_dp),
                                             DDATA(dsubgridErrorV_dv),
                                             DDATA(subgridErrorW),
                                             DDATA(dsubgridErrorW_dp),
                                             DDATA(dsubgridErrorW_dw));
   Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorStokes2D_GLS_velocity_pressure(PyObject* self,
                                                                 PyObject* args)
{
  PyObject *elementDiameter,
    *a,
    *pdeResidualP,
    *dpdeResidualP_du,
    *dpdeResidualP_dv,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_dv,
    *subgridErrorP,
    *dsubgridErrorP_du,
    *dsubgridErrorP_dv,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_dv;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOOOOOOOO",
                       &elementDiameter,
                       &a,
                       &pdeResidualP,
                       &dpdeResidualP_du,
                       &dpdeResidualP_dv,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_dv,
                       &subgridErrorP,
                       &dsubgridErrorP_du,
                       &dsubgridErrorP_dv,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_dv))
    return NULL;
  calculateSubgridErrorStokes2D_GLS_velocity_pressure(SHAPE(dsubgridErrorP_du)[0],
                                                      SHAPE(dsubgridErrorP_du)[1],
                                                      SHAPE(dsubgridErrorP_du)[2],
                                                      2,
                                                      DDATA(elementDiameter),
                                                      DDATA(a),
                                                      DDATA(pdeResidualP),
                                                      DDATA(dpdeResidualP_du),
                                                      DDATA(dpdeResidualP_dv),
                                                      DDATA(pdeResidualU),
                                                      DDATA(dpdeResidualU_dp),
                                                      DDATA(dpdeResidualU_du),
                                                      DDATA(pdeResidualV),
                                                      DDATA(dpdeResidualV_dp),
                                                      DDATA(dpdeResidualV_dv),
                                                      DDATA(subgridErrorP),
                                                      DDATA(dsubgridErrorP_du),
                                                      DDATA(dsubgridErrorP_dv),
                                                      DDATA(subgridErrorU),
                                                      DDATA(dsubgridErrorU_dp),
                                                      DDATA(dsubgridErrorU_du),
                                                      DDATA(subgridErrorV),
                                                      DDATA(dsubgridErrorV_dp),
                                                      DDATA(dsubgridErrorV_dv));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorStokes2D_GLS_velocity_pressure_sd(PyObject* self,
                                                                 PyObject* args)
{
  PyObject *elementDiameter,
    *a,
    *pdeResidualP,
    *dpdeResidualP_du,
    *dpdeResidualP_dv,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_dv,
    *subgridErrorP,
    *dsubgridErrorP_du,
    *dsubgridErrorP_dv,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_dv;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOOOOOOOO",
                       &elementDiameter,
                       &a,
                       &pdeResidualP,
                       &dpdeResidualP_du,
                       &dpdeResidualP_dv,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_dv,
                       &subgridErrorP,
                       &dsubgridErrorP_du,
                       &dsubgridErrorP_dv,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_dv))
    return NULL;
  calculateSubgridErrorStokes2D_GLS_velocity_pressure_sd(SHAPE(dsubgridErrorP_du)[0],
                                                      SHAPE(dsubgridErrorP_du)[1],
                                                      SHAPE(dsubgridErrorP_du)[2],
                                                      2,
                                                      DDATA(elementDiameter),
                                                      DDATA(a),
                                                      DDATA(pdeResidualP),
                                                      DDATA(dpdeResidualP_du),
                                                      DDATA(dpdeResidualP_dv),
                                                      DDATA(pdeResidualU),
                                                      DDATA(dpdeResidualU_dp),
                                                      DDATA(dpdeResidualU_du),
                                                      DDATA(pdeResidualV),
                                                      DDATA(dpdeResidualV_dp),
                                                      DDATA(dpdeResidualV_dv),
                                                      DDATA(subgridErrorP),
                                                      DDATA(dsubgridErrorP_du),
                                                      DDATA(dsubgridErrorP_dv),
                                                      DDATA(subgridErrorU),
                                                      DDATA(dsubgridErrorU_dp),
                                                      DDATA(dsubgridErrorU_du),
                                                      DDATA(subgridErrorV),
                                                      DDATA(dsubgridErrorV_dp),
                                                      DDATA(dsubgridErrorV_dv));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorStokes3D_GLS_velocity_pressure(PyObject* self,
                                                                 PyObject* args)
{
  PyObject *elementDiameter,
    *a,
    *pdeResidualP,
    *dpdeResidualP_du,
    *dpdeResidualP_dv,
    *dpdeResidualP_dw,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_dv,
    *pdeResidualW,
    *dpdeResidualW_dp,
    *dpdeResidualW_dw,
    *subgridErrorP,
    *dsubgridErrorP_du,
    *dsubgridErrorP_dv,
    *dsubgridErrorP_dw,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_dv,
    *subgridErrorW,
    *dsubgridErrorW_dp,
    *dsubgridErrorW_dw;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOOOOOOOOOOOOOOOOO",
                       &elementDiameter,
                       &a,
                       &pdeResidualP,
                       &dpdeResidualP_du,
                       &dpdeResidualP_dv,
                       &dpdeResidualP_dw,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_dv,
                       &pdeResidualW,
                       &dpdeResidualW_dp,
                       &dpdeResidualW_dw,
                       &subgridErrorP,
                       &dsubgridErrorP_du,
                       &dsubgridErrorP_dv,
                       &dsubgridErrorP_dw,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_dv,
                       &subgridErrorW,
                       &dsubgridErrorW_dp,
                       &dsubgridErrorW_dw))
    return NULL;
  calculateSubgridErrorStokes3D_GLS_velocity_pressure(SHAPE(dsubgridErrorP_du)[0],
                                                      SHAPE(dsubgridErrorP_du)[1],
                                                      SHAPE(dsubgridErrorP_du)[2],
                                                      3,
                                                      DDATA(elementDiameter),
                                                      DDATA(a),
                                                      DDATA(pdeResidualP),
                                                      DDATA(dpdeResidualP_du),
                                                      DDATA(dpdeResidualP_dv),
                                                      DDATA(dpdeResidualP_dw),
                                                      DDATA(pdeResidualU),
                                                      DDATA(dpdeResidualU_dp),
                                                      DDATA(dpdeResidualU_du),
                                                      DDATA(pdeResidualV),
                                                      DDATA(dpdeResidualV_dp),
                                                      DDATA(dpdeResidualV_dv),
                                                      DDATA(pdeResidualW),
                                                      DDATA(dpdeResidualW_dp),
                                                      DDATA(dpdeResidualW_dw),
                                                      DDATA(subgridErrorP),
                                                      DDATA(dsubgridErrorP_du),
                                                      DDATA(dsubgridErrorP_dv),
                                                      DDATA(dsubgridErrorP_dw),
                                                      DDATA(subgridErrorU),
                                                      DDATA(dsubgridErrorU_dp),
                                                      DDATA(dsubgridErrorU_du),
                                                      DDATA(subgridErrorV),
                                                      DDATA(dsubgridErrorV_dp),
                                                      DDATA(dsubgridErrorV_dv),
                                                      DDATA(subgridErrorW),
                                                      DDATA(dsubgridErrorW_dp),
                                                      DDATA(dsubgridErrorW_dw));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorStokes3D_GLS_velocity_pressure_sd(PyObject* self,
                                                                 PyObject* args)
{
  PyObject *elementDiameter,
    *a,
    *pdeResidualP,
    *dpdeResidualP_du,
    *dpdeResidualP_dv,
    *dpdeResidualP_dw,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_dv,
    *pdeResidualW,
    *dpdeResidualW_dp,
    *dpdeResidualW_dw,
    *subgridErrorP,
    *dsubgridErrorP_du,
    *dsubgridErrorP_dv,
    *dsubgridErrorP_dw,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_dv,
    *subgridErrorW,
    *dsubgridErrorW_dp,
    *dsubgridErrorW_dw;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOOOOOOOOOOOOOOOOO",
                       &elementDiameter,
                       &a,
                       &pdeResidualP,
                       &dpdeResidualP_du,
                       &dpdeResidualP_dv,
                       &dpdeResidualP_dw,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_dv,
                       &pdeResidualW,
                       &dpdeResidualW_dp,
                       &dpdeResidualW_dw,
                       &subgridErrorP,
                       &dsubgridErrorP_du,
                       &dsubgridErrorP_dv,
                       &dsubgridErrorP_dw,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_dv,
                       &subgridErrorW,
                       &dsubgridErrorW_dp,
                       &dsubgridErrorW_dw))
    return NULL;
  calculateSubgridErrorStokes3D_GLS_velocity_pressure_sd(SHAPE(dsubgridErrorP_du)[0],
                                                      SHAPE(dsubgridErrorP_du)[1],
                                                      SHAPE(dsubgridErrorP_du)[2],
                                                      3,
                                                      DDATA(elementDiameter),
                                                      DDATA(a),
                                                      DDATA(pdeResidualP),
                                                      DDATA(dpdeResidualP_du),
                                                      DDATA(dpdeResidualP_dv),
                                                      DDATA(dpdeResidualP_dw),
                                                      DDATA(pdeResidualU),
                                                      DDATA(dpdeResidualU_dp),
                                                      DDATA(dpdeResidualU_du),
                                                      DDATA(pdeResidualV),
                                                      DDATA(dpdeResidualV_dp),
                                                      DDATA(dpdeResidualV_dv),
                                                      DDATA(pdeResidualW),
                                                      DDATA(dpdeResidualW_dp),
                                                      DDATA(dpdeResidualW_dw),
                                                      DDATA(subgridErrorP),
                                                      DDATA(dsubgridErrorP_du),
                                                      DDATA(dsubgridErrorP_dv),
                                                      DDATA(dsubgridErrorP_dw),
                                                      DDATA(subgridErrorU),
                                                      DDATA(dsubgridErrorU_dp),
                                                      DDATA(dsubgridErrorU_du),
                                                      DDATA(subgridErrorV),
                                                      DDATA(dsubgridErrorV_dp),
                                                      DDATA(dsubgridErrorV_dv),
                                                      DDATA(subgridErrorW),
                                                      DDATA(dsubgridErrorW_dp),
                                                      DDATA(dsubgridErrorW_dw));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorNavierStokes2D_GLS_tau(PyObject* self,
                                                         PyObject* args)
{
  double hFactor;
  PyObject *elementDiameter,
    *dmt,
    *dm,
    *f,
    *a,
    *tau0,
    *tau1,
    *cfl;
  if(!PyArg_ParseTuple(args,"dOOOOOOOO",
                       &hFactor,
                       &elementDiameter,
                       &dmt,
                       &dm,
                       &f,
                       &a,
                       &tau0,
                       &tau1,
                       &cfl))
    return NULL;
  calculateSubgridErrorNavierStokes2D_GLS_tau(SHAPE(f)[0],
                                              SHAPE(f)[1],
                                              SHAPE(f)[2],
                                              hFactor,
                                              DDATA(elementDiameter),
                                              DDATA(dmt),
                                              DDATA(dm),
                                              DDATA(f),
                                              DDATA(a),
                                              DDATA(tau0),
                                              DDATA(tau1),
                                              DDATA(cfl));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorNavierStokes2D_GLS_tau_sd(PyObject* self,
                                                         PyObject* args)
{
  double hFactor;
  PyObject *elementDiameter,
    *dmt,
    *dm,
    *f,
    *a,
    *tau0,
    *tau1,
    *cfl;
  if(!PyArg_ParseTuple(args,"dOOOOOOOO",
                       &hFactor,
                       &elementDiameter,
                       &dmt,
                       &dm,
                       &f,
                       &a,
                       &tau0,
                       &tau1,
                       &cfl))
    return NULL;
  calculateSubgridErrorNavierStokes2D_GLS_tau_sd(SHAPE(f)[0],
                                              SHAPE(f)[1],
                                              SHAPE(f)[2],
                                              hFactor,
                                              DDATA(elementDiameter),
                                              DDATA(dmt),
                                              DDATA(dm),
                                              DDATA(f),
                                              DDATA(a),
                                              DDATA(tau0),
                                              DDATA(tau1),
                                              DDATA(cfl));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorNavierStokes2D_generic_tau(PyObject* self,
                                                         PyObject* args)
{
  PyObject *inverseJ,
    *dmt,
    *dm,
    *f,
    *a,
    *tau0,
    *tau1,
    *cfl;
  if(!PyArg_ParseTuple(args,"OOOOOOOO",
                       &inverseJ,
                       &dmt,
                       &dm,
                       &f,
                       &a,
                       &tau0,
                       &tau1,
                       &cfl))
    return NULL;
  calculateSubgridErrorNavierStokes2D_generic_tau(SHAPE(f)[0],
                                                  SHAPE(f)[1],
                                                  SHAPE(f)[2],
                                                  DDATA(inverseJ),
                                                  DDATA(dmt),
                                                  DDATA(dm),
                                                  DDATA(f),
                                                  DDATA(a),
                                                  DDATA(tau0),
                                                  DDATA(tau1),
                                                  DDATA(cfl));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorNavierStokes2D_generic_tau_sd(PyObject* self,
                                                         PyObject* args)
{
  PyObject *inverseJ,
    *dmt,
    *dm,
    *f,
    *a,
    *tau0,
    *tau1,
    *cfl;
  if(!PyArg_ParseTuple(args,"OOOOOOOO",
                       &inverseJ,
                       &dmt,
                       &dm,
                       &f,
                       &a,
                       &tau0,
                       &tau1,
                       &cfl))
    return NULL;
  calculateSubgridErrorNavierStokes2D_generic_tau_sd(SHAPE(f)[0],
                                                  SHAPE(f)[1],
                                                  SHAPE(f)[2],
                                                  DDATA(inverseJ),
                                                  DDATA(dmt),
                                                  DDATA(dm),
                                                  DDATA(f),
                                                  DDATA(a),
                                                  DDATA(tau0),
                                                  DDATA(tau1),
                                                  DDATA(cfl));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorNavierStokes2D_generic_withBodyForce_tau(PyObject* self,
									   PyObject* args)
{
  PyObject *inverseJ,
    *dmt,
    *dm,
    *f,
    *a,
    *dr,
    *tau0,
    *tau1,
    *cfl;
  if(!PyArg_ParseTuple(args,"OOOOOOOOO",
                       &inverseJ,
                       &dmt,
                       &dm,
                       &f,
                       &a,
		       &dr,
                       &tau0,
                       &tau1,
                       &cfl))
    return NULL;
  calculateSubgridErrorNavierStokes2D_generic_withBodyForce_tau(SHAPE(f)[0],
								SHAPE(f)[1],
								SHAPE(f)[2],
								DDATA(inverseJ),
								DDATA(dmt),
								DDATA(dm),
								DDATA(f),
								DDATA(a),
								DDATA(dr),
								DDATA(tau0),
								DDATA(tau1),
								DDATA(cfl));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorNavierStokes2D_generic_withBodyForce_tau_sd(PyObject* self,
									   PyObject* args)
{
  PyObject *inverseJ,
    *dmt,
    *dm,
    *f,
    *a,
    *dr,
    *tau0,
    *tau1,
    *cfl;
  if(!PyArg_ParseTuple(args,"OOOOOOOOO",
                       &inverseJ,
                       &dmt,
                       &dm,
                       &f,
                       &a,
		       &dr,
                       &tau0,
                       &tau1,
                       &cfl))
    return NULL;
  calculateSubgridErrorNavierStokes2D_generic_withBodyForce_tau_sd(SHAPE(f)[0],
								SHAPE(f)[1],
								SHAPE(f)[2],
								DDATA(inverseJ),
								DDATA(dmt),
								DDATA(dm),
								DDATA(f),
								DDATA(a),
								DDATA(dr),
								DDATA(tau0),
								DDATA(tau1),
								DDATA(cfl));
  Py_INCREF(Py_None);
  return Py_None;
}



static PyObject*
csubgridErrorCalculateSubgridErrorNavierStokes2D_GLS_tauRes(PyObject* self,
                                                            PyObject* args)
{
  PyObject *tau0,
    *tau1,
    *pdeResidualP,
    *dpdeResidualP_du,
    *dpdeResidualP_dv,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *dpdeResidualU_dv,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_du,
    *dpdeResidualV_dv,
    *subgridErrorP,
    *dsubgridErrorP_du,
    *dsubgridErrorP_dv,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *dsubgridErrorU_dv,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_du,
    *dsubgridErrorV_dv;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOOOOOOOOOOOO",
                       &tau0,
                       &tau1,
                       &pdeResidualP,
                       &dpdeResidualP_du,
                       &dpdeResidualP_dv,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &dpdeResidualU_dv,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_du,
                       &dpdeResidualV_dv,
                       &subgridErrorP,
                       &dsubgridErrorP_du,
                       &dsubgridErrorP_dv,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &dsubgridErrorU_dv,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_du,
                       &dsubgridErrorV_dv))
    return NULL;
  calculateSubgridErrorNavierStokes2D_GLS_tauRes(SHAPE(dsubgridErrorP_du)[0],
                                                 SHAPE(dsubgridErrorP_du)[1],
                                                 SHAPE(dsubgridErrorP_du)[2],
                                                 2,
                                                 DDATA(tau0),
                                                 DDATA(tau1),
                                                 DDATA(pdeResidualP),
                                                 DDATA(dpdeResidualP_du),
                                                 DDATA(dpdeResidualP_dv),
                                                 DDATA(pdeResidualU),
                                                 DDATA(dpdeResidualU_dp),
                                                 DDATA(dpdeResidualU_du),
                                                 DDATA(dpdeResidualU_dv),
                                                 DDATA(pdeResidualV),
                                                 DDATA(dpdeResidualV_dp),
                                                 DDATA(dpdeResidualV_du),
                                                 DDATA(dpdeResidualV_dv),
                                                 DDATA(subgridErrorP),
                                                 DDATA(dsubgridErrorP_du),
                                                 DDATA(dsubgridErrorP_dv),
                                                 DDATA(subgridErrorU),
                                                 DDATA(dsubgridErrorU_dp),
                                                 DDATA(dsubgridErrorU_du),
                                                 DDATA(dsubgridErrorU_dv),
                                                 DDATA(subgridErrorV),
                                                 DDATA(dsubgridErrorV_dp),
                                                 DDATA(dsubgridErrorV_du),
                                                 DDATA(dsubgridErrorV_dv));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorNavierStokes3D_GLS_tauRes(PyObject* self,
                                                            PyObject* args)
{
  PyObject *tau0,
    *tau1,
    *pdeResidualP,
    *dpdeResidualP_du,
    *dpdeResidualP_dv,
    *dpdeResidualP_dw,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *dpdeResidualU_dv,
    *dpdeResidualU_dw,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_du,
    *dpdeResidualV_dv,
    *dpdeResidualV_dw,
    *pdeResidualW,
    *dpdeResidualW_dp,
    *dpdeResidualW_du,
    *dpdeResidualW_dv,
    *dpdeResidualW_dw,
    *subgridErrorP,
    *dsubgridErrorP_du,
    *dsubgridErrorP_dv,
    *dsubgridErrorP_dw,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *dsubgridErrorU_dv,
    *dsubgridErrorU_dw,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_du,
    *dsubgridErrorV_dv,
    *dsubgridErrorV_dw,
    *subgridErrorW,
    *dsubgridErrorW_dp,
    *dsubgridErrorW_du,
    *dsubgridErrorW_dv,
    *dsubgridErrorW_dw;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOOO",
                       &tau0,
                       &tau1,
                       &pdeResidualP,
                       &dpdeResidualP_du,
                       &dpdeResidualP_dv,
                       &dpdeResidualP_dw,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &dpdeResidualU_dv,
                       &dpdeResidualU_dw,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_du,
                       &dpdeResidualV_dv,
                       &dpdeResidualV_dw,
                       &pdeResidualW,
                       &dpdeResidualW_dp,
                       &dpdeResidualW_du,
                       &dpdeResidualW_dv,
                       &dpdeResidualW_dw,
                       &subgridErrorP,
                       &dsubgridErrorP_du,
                       &dsubgridErrorP_dv,
                       &dsubgridErrorP_dw,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &dsubgridErrorU_dv,
                       &dsubgridErrorU_dw,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_du,
                       &dsubgridErrorV_dv,
                       &dsubgridErrorV_dw,
                       &subgridErrorW,
                       &dsubgridErrorW_dp,
                       &dsubgridErrorW_du,
                       &dsubgridErrorW_dv,
                       &dsubgridErrorW_dw))
    return NULL;
  calculateSubgridErrorNavierStokes3D_GLS_tauRes(SHAPE(dsubgridErrorP_du)[0],
                                                 SHAPE(dsubgridErrorP_du)[1],
                                                 SHAPE(dsubgridErrorP_du)[2],
                                                 3,
                                                 DDATA(tau0),
                                                 DDATA(tau1),
                                                 DDATA(pdeResidualP),
                                                 DDATA(dpdeResidualP_du),
                                                 DDATA(dpdeResidualP_dv),
                                                 DDATA(dpdeResidualP_dw),
                                                 DDATA(pdeResidualU),
                                                 DDATA(dpdeResidualU_dp),
                                                 DDATA(dpdeResidualU_du),
                                                 DDATA(dpdeResidualU_dv),
                                                 DDATA(dpdeResidualU_dw),
                                                 DDATA(pdeResidualV),
                                                 DDATA(dpdeResidualV_dp),
                                                 DDATA(dpdeResidualV_du),
                                                 DDATA(dpdeResidualV_dv),
                                                 DDATA(dpdeResidualV_dw),
                                                 DDATA(pdeResidualW),
                                                 DDATA(dpdeResidualW_dp),
                                                 DDATA(dpdeResidualW_du),
                                                 DDATA(dpdeResidualW_dv),
                                                 DDATA(dpdeResidualW_dw),
                                                 DDATA(subgridErrorP),
                                                 DDATA(dsubgridErrorP_du),
                                                 DDATA(dsubgridErrorP_dv),
                                                 DDATA(dsubgridErrorP_dw),
                                                 DDATA(subgridErrorU),
                                                 DDATA(dsubgridErrorU_dp),
                                                 DDATA(dsubgridErrorU_du),
                                                 DDATA(dsubgridErrorU_dv),
                                                 DDATA(dsubgridErrorU_dw),
                                                 DDATA(subgridErrorV),
                                                 DDATA(dsubgridErrorV_dp),
                                                 DDATA(dsubgridErrorV_du),
                                                 DDATA(dsubgridErrorV_dv),
                                                 DDATA(dsubgridErrorV_dw),
                                                 DDATA(subgridErrorW),
                                                 DDATA(dsubgridErrorW_dp),
                                                 DDATA(dsubgridErrorW_du),
                                                 DDATA(dsubgridErrorW_dv),
                                                 DDATA(dsubgridErrorW_dw));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorStokes_GLS_tau(PyObject* self,
                                                 PyObject* args)
{
  PyObject *elementDiameter,
    *dH,
    *a,
    *tau0,
    *tau1;
  if(!PyArg_ParseTuple(args,"OOOOO",
                       &elementDiameter,
                       &dH,
                       &a,
                       &tau0,
                       &tau1))
    return NULL;
  calculateSubgridErrorStokes_GLS_tau(SHAPE(dH)[0],
                                      SHAPE(dH)[1],
                                      SHAPE(dH)[2],
                                      DDATA(elementDiameter),
                                      DDATA(dH),
                                      DDATA(a),
                                      DDATA(tau0),
                                      DDATA(tau1));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorStokes_GLS_tau_sd(PyObject* self,
                                                    PyObject* args)
{
  PyObject *elementDiameter,
    *dH,
    *a,
    *tau0,
    *tau1;
  if(!PyArg_ParseTuple(args,"OOOOO",
                       &elementDiameter,
                       &dH,
                       &a,
                       &tau0,
                       &tau1))
    return NULL;
  calculateSubgridErrorStokes_GLS_tau_sd(SHAPE(dH)[0],
                                         SHAPE(dH)[1],
                                         SHAPE(dH)[2],
                                         DDATA(elementDiameter),
                                         DDATA(dH),
                                         DDATA(a),
                                         DDATA(tau0),
                                         DDATA(tau1));
  Py_INCREF(Py_None);
  return Py_None;
}


static PyObject*
csubgridErrorCalculateSubgridErrorStokes2D_GLS_tauRes(PyObject* self,
                                                      PyObject* args)
{
  PyObject *tau0,
    *tau1,
    *pdeResidualP,
    *dpdeResidualP_du,
    *dpdeResidualP_dv,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_dv,
    *subgridErrorP,
    *dsubgridErrorP_du,
    *dsubgridErrorP_dv,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_dv;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOOOOOOOO",
                       &tau0,
                       &tau1,
                       &pdeResidualP,
                       &dpdeResidualP_du,
                       &dpdeResidualP_dv,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_dv,
                       &subgridErrorP,
                       &dsubgridErrorP_du,
                       &dsubgridErrorP_dv,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_dv))
    return NULL;
  calculateSubgridErrorStokes2D_GLS_tauRes(SHAPE(dsubgridErrorP_du)[0],
                                           SHAPE(dsubgridErrorP_du)[1],
                                           SHAPE(dsubgridErrorP_du)[2],
                                           2,
                                           DDATA(tau0),
                                           DDATA(tau1),
                                           DDATA(pdeResidualP),
                                           DDATA(dpdeResidualP_du),
                                           DDATA(dpdeResidualP_dv),
                                           DDATA(pdeResidualU),
                                           DDATA(dpdeResidualU_dp),
                                           DDATA(dpdeResidualU_du),
                                           DDATA(pdeResidualV),
                                           DDATA(dpdeResidualV_dp),
                                           DDATA(dpdeResidualV_dv),
                                           DDATA(subgridErrorP),
                                           DDATA(dsubgridErrorP_du),
                                           DDATA(dsubgridErrorP_dv),
                                           DDATA(subgridErrorU),
                                           DDATA(dsubgridErrorU_dp),
                                           DDATA(dsubgridErrorU_du),
                                           DDATA(subgridErrorV),
                                           DDATA(dsubgridErrorV_dp),
                                           DDATA(dsubgridErrorV_dv));
  Py_INCREF(Py_None);
  return Py_None;
}
static PyObject*
csubgridErrorCalculateSubgridErrorStokes3D_GLS_tauRes(PyObject* self,
                                                      PyObject* args)
{
  PyObject *tau0,
    *tau1,
    *pdeResidualP,
    *dpdeResidualP_du,
    *dpdeResidualP_dv,
    *dpdeResidualP_dw,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_dv,
    *pdeResidualW,
    *dpdeResidualW_dp,
    *dpdeResidualW_dw,
    *subgridErrorP,
    *dsubgridErrorP_du,
    *dsubgridErrorP_dv,
    *dsubgridErrorP_dw,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_dv,
    *subgridErrorW,
    *dsubgridErrorW_dp,
    *dsubgridErrorW_dw;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOOOOOOOOOOOOOOOO",
                       &tau0,
                       &tau1,
                       &pdeResidualP,
                       &dpdeResidualP_du,
                       &dpdeResidualP_dv,
                       &dpdeResidualP_dw,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_dv,
                       &pdeResidualW,
                       &dpdeResidualW_dp,
                       &dpdeResidualW_dw,
                       &subgridErrorP,
                       &dsubgridErrorP_du,
                       &dsubgridErrorP_dv,
                       &dsubgridErrorP_dw,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_dv,
                       &subgridErrorW,
                       &dsubgridErrorW_dp,
                       &dsubgridErrorW_dw))
    return NULL;
  calculateSubgridErrorStokes3D_GLS_tauRes(SHAPE(dsubgridErrorP_du)[0],
                                           SHAPE(dsubgridErrorP_du)[1],
                                           SHAPE(dsubgridErrorP_du)[2],
                                           3,
                                           DDATA(tau0),
                                           DDATA(tau1),
                                           DDATA(pdeResidualP),
                                           DDATA(dpdeResidualP_du),
                                           DDATA(dpdeResidualP_dv),
                                           DDATA(dpdeResidualP_dw),
                                           DDATA(pdeResidualU),
                                           DDATA(dpdeResidualU_dp),
                                           DDATA(dpdeResidualU_du),
                                           DDATA(pdeResidualV),
                                           DDATA(dpdeResidualV_dp),
                                           DDATA(dpdeResidualV_dv),
                                           DDATA(pdeResidualW),
                                           DDATA(dpdeResidualW_dp),
                                           DDATA(dpdeResidualW_dw),
                                           DDATA(subgridErrorP),
                                           DDATA(dsubgridErrorP_du),
                                           DDATA(dsubgridErrorP_dv),
                                           DDATA(dsubgridErrorP_dw),
                                           DDATA(subgridErrorU),
                                           DDATA(dsubgridErrorU_dp),
                                           DDATA(dsubgridErrorU_du),
                                           DDATA(subgridErrorV),
                                           DDATA(dsubgridErrorV_dp),
                                           DDATA(dsubgridErrorV_dv),
					   DDATA(subgridErrorW),
                                           DDATA(dsubgridErrorW_dp),
                                           DDATA(dsubgridErrorW_dw));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorNavierStokes3D_GLS_velocity_pressure(PyObject* self,
                                                                       PyObject* args)
{
  PyObject *elementDiameter,
    *dm,
    *f,
    *a,
    *pdeResidualP,
    *dpdeResidualP_du,
    *dpdeResidualP_dv,
    *dpdeResidualP_dw,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_dv,
    *pdeResidualW,
    *dpdeResidualW_dp,
    *dpdeResidualW_dw,
    *subgridErrorP,
    *dsubgridErrorP_du,
    *dsubgridErrorP_dv,
    *dsubgridErrorP_dw,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_dv,
    *subgridErrorW,
    *dsubgridErrorW_dp,
    *dsubgridErrorW_dw;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOOOOOOOOOOOOOOOOOO",
                       &elementDiameter,
                       &dm,
                       &f,
                       &a,
                       &pdeResidualP,
                       &dpdeResidualP_du,
                       &dpdeResidualP_dv,
                       &dpdeResidualP_dw,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_dv,
                       &pdeResidualW,
                       &dpdeResidualW_dp,
                       &dpdeResidualW_dw,
                       &subgridErrorP,
                       &dsubgridErrorP_du,
                       &dsubgridErrorP_dv,
                       &dsubgridErrorP_dw,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_dv,
                       &subgridErrorW,
                       &dsubgridErrorW_dp,
                       &dsubgridErrorW_dw))
    return NULL;
  calculateSubgridErrorNavierStokes3D_GLS_velocity_pressure(SHAPE(dsubgridErrorP_du)[0],
                                                            SHAPE(dsubgridErrorP_du)[1],
                                                            SHAPE(dsubgridErrorP_du)[2],
                                                            3,
                                                            DDATA(elementDiameter),
                                                            DDATA(dm),
                                                            DDATA(f),
                                                            DDATA(a),
                                                            DDATA(pdeResidualP),
                                                            DDATA(dpdeResidualP_du),
                                                            DDATA(dpdeResidualP_dv),
                                                            DDATA(dpdeResidualP_dw),
                                                            DDATA(pdeResidualU),
                                                            DDATA(dpdeResidualU_dp),
                                                            DDATA(dpdeResidualU_du),
                                                            DDATA(pdeResidualV),
                                                            DDATA(dpdeResidualV_dp),
                                                            DDATA(dpdeResidualV_dv),
                                                            DDATA(pdeResidualW),
                                                            DDATA(dpdeResidualW_dp),
                                                            DDATA(dpdeResidualW_dw),
                                                            DDATA(subgridErrorP),
                                                            DDATA(dsubgridErrorP_du),
                                                            DDATA(dsubgridErrorP_dv),
                                                            DDATA(dsubgridErrorP_dw),
                                                            DDATA(subgridErrorU),
                                                            DDATA(dsubgridErrorU_dp),
                                                            DDATA(dsubgridErrorU_du),
                                                            DDATA(subgridErrorV),
                                                            DDATA(dsubgridErrorV_dp),
                                                            DDATA(dsubgridErrorV_dv),
                                                            DDATA(subgridErrorW),
                                                            DDATA(dsubgridErrorW_dp),
                                                            DDATA(dsubgridErrorW_dw));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorNavierStokes3D_GLS_velocity_pressure_sd(PyObject* self,
                                                                       PyObject* args)
{
  PyObject *elementDiameter,
    *dm,
    *f,
    *a,
    *pdeResidualP,
    *dpdeResidualP_du,
    *dpdeResidualP_dv,
    *dpdeResidualP_dw,
    *pdeResidualU,
    *dpdeResidualU_dp,
    *dpdeResidualU_du,
    *pdeResidualV,
    *dpdeResidualV_dp,
    *dpdeResidualV_dv,
    *pdeResidualW,
    *dpdeResidualW_dp,
    *dpdeResidualW_dw,
    *subgridErrorP,
    *dsubgridErrorP_du,
    *dsubgridErrorP_dv,
    *dsubgridErrorP_dw,
    *subgridErrorU,
    *dsubgridErrorU_dp,
    *dsubgridErrorU_du,
    *subgridErrorV,
    *dsubgridErrorV_dp,
    *dsubgridErrorV_dv,
    *subgridErrorW,
    *dsubgridErrorW_dp,
    *dsubgridErrorW_dw;
  if(!PyArg_ParseTuple(args,"OOOOOOOOOOOOOOOOOOOOOOOOOOOOOO",
                       &elementDiameter,
                       &dm,
                       &f,
                       &a,
                       &pdeResidualP,
                       &dpdeResidualP_du,
                       &dpdeResidualP_dv,
                       &dpdeResidualP_dw,
                       &pdeResidualU,
                       &dpdeResidualU_dp,
                       &dpdeResidualU_du,
                       &pdeResidualV,
                       &dpdeResidualV_dp,
                       &dpdeResidualV_dv,
                       &pdeResidualW,
                       &dpdeResidualW_dp,
                       &dpdeResidualW_dw,
                       &subgridErrorP,
                       &dsubgridErrorP_du,
                       &dsubgridErrorP_dv,
                       &dsubgridErrorP_dw,
                       &subgridErrorU,
                       &dsubgridErrorU_dp,
                       &dsubgridErrorU_du,
                       &subgridErrorV,
                       &dsubgridErrorV_dp,
                       &dsubgridErrorV_dv,
                       &subgridErrorW,
                       &dsubgridErrorW_dp,
                       &dsubgridErrorW_dw))
    return NULL;
  calculateSubgridErrorNavierStokes3D_GLS_velocity_pressure_sd(SHAPE(dsubgridErrorP_du)[0],
                                                            SHAPE(dsubgridErrorP_du)[1],
                                                            SHAPE(dsubgridErrorP_du)[2],
                                                            3,
                                                            DDATA(elementDiameter),
                                                            DDATA(dm),
                                                            DDATA(f),
                                                            DDATA(a),
                                                            DDATA(pdeResidualP),
                                                            DDATA(dpdeResidualP_du),
                                                            DDATA(dpdeResidualP_dv),
                                                            DDATA(dpdeResidualP_dw),
                                                            DDATA(pdeResidualU),
                                                            DDATA(dpdeResidualU_dp),
                                                            DDATA(dpdeResidualU_du),
                                                            DDATA(pdeResidualV),
                                                            DDATA(dpdeResidualV_dp),
                                                            DDATA(dpdeResidualV_dv),
                                                            DDATA(pdeResidualW),
                                                            DDATA(dpdeResidualW_dp),
                                                            DDATA(dpdeResidualW_dw),
                                                            DDATA(subgridErrorP),
                                                            DDATA(dsubgridErrorP_du),
                                                            DDATA(dsubgridErrorP_dv),
                                                            DDATA(dsubgridErrorP_dw),
                                                            DDATA(subgridErrorU),
                                                            DDATA(dsubgridErrorU_dp),
                                                            DDATA(dsubgridErrorU_du),
                                                            DDATA(subgridErrorV),
                                                            DDATA(dsubgridErrorV_dp),
                                                            DDATA(dsubgridErrorV_dv),
                                                            DDATA(subgridErrorW),
                                                            DDATA(dsubgridErrorW_dp),
                                                            DDATA(dsubgridErrorW_dw));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorShallowWater1D(PyObject* self,
                                                 PyObject* args)
{
  double g;
  PyObject *elementDiameter,
    *h,
    *hu,
    *cfl_1,
    *cfl_2;
  if(!PyArg_ParseTuple(args,"dOOOOO",
                       &g,
                       &elementDiameter,
                       &h,
                       &hu,
                       &cfl_1,
                       &cfl_2))
    return NULL;
  calculateSubgridErrorShallowWater1D(SHAPE(h)[0],
                                      SHAPE(h)[1],
                                      g,
                                      DDATA(elementDiameter),
                                      DDATA(h),
                                      DDATA(hu),
                                      DDATA(cfl_1),
                                      DDATA(cfl_2));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridErrorShallowWater2D(PyObject* self,
                                                 PyObject* args)
{
  double g;
  PyObject *elementDiameter,
    *h,
    *hu,
    *hv,
    *cfl_1,
    *cfl_2,
    *cfl_3;
  if(!PyArg_ParseTuple(args,"dOOOOOOO",
                       &g,
                       &elementDiameter,
                       &h,
                       &hu,
                       &hv,
                       &cfl_1,
                       &cfl_2,
                       &cfl_3))
    return NULL;
  calculateSubgridErrorShallowWater2D(SHAPE(h)[0],
                                      SHAPE(h)[1],
                                      g,
                                      DDATA(elementDiameter),
                                      DDATA(h),
                                      DDATA(hu),
                                      DDATA(hv),
                                      DDATA(cfl_1),
                                      DDATA(cfl_2),
                                      DDATA(cfl_3));
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject*
csubgridErrorCalculateSubgridError_Harari_tau_sd(PyObject* self,
                                                 PyObject* args)
{
  PyObject *elementDiameter,**a,*tau,*rowptr,*colind;
  int nSpace;
  double dt;
  if(!PyArg_ParseTuple(args,"idOOOOO",
		       &nSpace,
		       &dt,
		       &rowptr,
		       &colind,
                       &elementDiameter,
                       &a,
                       &tau))
    return NULL;
  calculateSubgridError_Harari_tau_sd(SHAPE(tau)[0],
				      SHAPE(tau)[1],
				      nSpace,
				      dt,
				      IDATA(rowptr),
				      IDATA(colind),
				      DDATA(elementDiameter),
				      DDATA(a),
				      DDATA(tau));

  Py_INCREF(Py_None);
  return Py_None;
}

static PyMethodDef csubgridErrorMethods[] = {
  { "calculateSubgridError_tauRes", 
    csubgridErrorCalculateSubgridError_tauRes, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the scalar advection equation using formula 1"},
  { "calculateSubgridErrorGradient_tauRes", 
    csubgridErrorCalculateSubgridErrorGradient_tauRes, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the gradient stabilization term in a scalar advection equation"},
  { "calculateSubgridError_ADR_tau", 
    csubgridErrorCalculateSubgridError_ADR_tau, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the scalar ADR equation using formula 1"},
  { "calculateSubgridError_ADR_tau_sd", 
    csubgridErrorCalculateSubgridError_ADR_tau_sd, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the scalar ADR equation using formula 1"},
  { "calculateSubgridError_ADR_generic_tau", 
    csubgridErrorCalculateSubgridError_ADR_generic_tau, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the scalar ADR equation using formula 1"},
  { "calculateSubgridError_ADR_Sangalli_tau_sd", 
    csubgridErrorCalculateSubgridError_ADR_Sangalli_tau_sd, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the scalar ADR equation using formula 1"},
  { "calculateSubgridError_ADR_generic_tau_sd", 
    csubgridErrorCalculateSubgridError_ADR_generic_tau_sd, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the scalar ADR equation using formula 1"},
  { "calculateSubgridError_A_tau", 
    csubgridErrorCalculateSubgridError_A_tau, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the scalar ADR equation using formula 1"},
  { "calculateSubgridError_HJ_tau", 
    csubgridErrorCalculateSubgridError_HJ_tau, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the scalar ADR equation using formula 1"},
  { "calculateSubgridErrorStokes2D_1", 
    csubgridErrorCalculateSubgridErrorStokes2D_1, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D Stokes equation using formula 1"},
  { "calculateSubgridErrorStokes2D_1_sd", 
    csubgridErrorCalculateSubgridErrorStokes2D_1_sd, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D Stokes equation using formula 1"},
  { "calculateSubgridErrorStokes2D_GLS_velocity", 
    csubgridErrorCalculateSubgridErrorStokes2D_GLS_velocity, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D Stokes equation using GLS formula for velocity"},
  { "calculateSubgridErrorStokes2D_GLS_velocity_sd", 
    csubgridErrorCalculateSubgridErrorStokes2D_GLS_velocity_sd, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D Stokes equation using GLS formula for velocity"},
  { "calculateSubgridErrorStokes3D_GLS_velocity", 
    csubgridErrorCalculateSubgridErrorStokes3D_GLS_velocity, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 3D Stokes equation using GLS formula for velocity"},
  { "calculateSubgridErrorStokes3D_GLS_velocity_sd", 
    csubgridErrorCalculateSubgridErrorStokes3D_GLS_velocity_sd, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 3D Stokes equation using GLS formula for velocity"},
  { "calculateSubgridErrorStokes2D_GLS_velocity_pressure", 
    csubgridErrorCalculateSubgridErrorStokes2D_GLS_velocity_pressure, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D Stokes equation using GLS formula for velocity and pressure"},
  { "calculateSubgridErrorStokes2D_GLS_velocity_pressure_sd", 
    csubgridErrorCalculateSubgridErrorStokes2D_GLS_velocity_pressure_sd, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D Stokes equation using GLS formula for velocity and pressure"},
  { "calculateSubgridErrorStokes3D_GLS_velocity_pressure", 
    csubgridErrorCalculateSubgridErrorStokes3D_GLS_velocity_pressure, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 3D Stokes equation using GLS formula for velocity and pressure"},
  { "calculateSubgridErrorStokes3D_GLS_velocity_pressure_sd", 
    csubgridErrorCalculateSubgridErrorStokes3D_GLS_velocity_pressure_sd, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 3D Stokes equation using GLS formula for velocity and pressure"},
  { "calculateSubgridErrorNavierStokes2D_GLS_tau", 
    csubgridErrorCalculateSubgridErrorNavierStokes2D_GLS_tau, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D NavierStokes equation using GLS formula for velocity and pressure"},
  { "calculateSubgridErrorNavierStokes2D_GLS_tau_sd", 
    csubgridErrorCalculateSubgridErrorNavierStokes2D_GLS_tau_sd, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D NavierStokes equation using GLS formula for velocity and pressure"},
  { "calculateSubgridErrorNavierStokes2D_generic_tau", 
    csubgridErrorCalculateSubgridErrorNavierStokes2D_generic_tau, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D Stokes equation using what I think is a general formula for velocity and pressure tau"},
  { "calculateSubgridErrorNavierStokes2D_generic_tau_sd", 
    csubgridErrorCalculateSubgridErrorNavierStokes2D_generic_tau_sd, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D Stokes equation using what I think is a general formula for velocity and pressure tau"},
  { "calculateSubgridErrorNavierStokes2D_generic_withBodyForce_tau", 
    csubgridErrorCalculateSubgridErrorNavierStokes2D_generic_withBodyForce_tau, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D Stokes equation using what I think is a general formula for velocity and pressure tau"},
  { "calculateSubgridErrorNavierStokes2D_generic_withBodyForce_tau_sd", 
    csubgridErrorCalculateSubgridErrorNavierStokes2D_generic_withBodyForce_tau_sd, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D Stokes equation using what I think is a general formula for velocity and pressure tau"},
  { "calculateSubgridErrorNavierStokes2D_GLS_tauRes", 
    csubgridErrorCalculateSubgridErrorNavierStokes2D_GLS_tauRes, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D NavierStokes equation using GLS formula for velocity and pressure"},
  { "calculateSubgridErrorNavierStokes3D_GLS_tauRes", 
    csubgridErrorCalculateSubgridErrorNavierStokes3D_GLS_tauRes, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D NavierStokes equation using GLS formula for velocity and pressure"},
  { "calculateSubgridErrorStokes_GLS_tau", 
    csubgridErrorCalculateSubgridErrorStokes_GLS_tau, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the Stokes equation using GLS formula for velocity and pressure"},
  { "calculateSubgridErrorStokes_GLS_tau_sd", 
    csubgridErrorCalculateSubgridErrorStokes_GLS_tau_sd, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the Stokes equation using GLS formula for velocity and pressure"},
  { "calculateSubgridErrorStokes2D_GLS_tauRes", 
    csubgridErrorCalculateSubgridErrorStokes2D_GLS_tauRes, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D Stokes equation using GLS formula for velocity and pressure"},
  { "calculateSubgridErrorStokes3D_GLS_tauRes", 
    csubgridErrorCalculateSubgridErrorStokes3D_GLS_tauRes, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 3D Stokes equation using GLS formula for velocity and pressure"},
  { "calculateSubgridErrorNavierStokes3D_GLS_velocity_pressure", 
    csubgridErrorCalculateSubgridErrorNavierStokes3D_GLS_velocity_pressure, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 3D NavierStokes equation using GLS formula for velocity and pressure"},
  { "calculateSubgridErrorNavierStokes3D_GLS_velocity_pressure_sd", 
    csubgridErrorCalculateSubgridErrorNavierStokes3D_GLS_velocity_pressure_sd, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 3D NavierStokes equation using GLS formula for velocity and pressure"},
  { "calculateSubgridErrorShallowWater1D", 
    csubgridErrorCalculateSubgridErrorShallowWater1D, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 1D shallow water equations"},
  { "calculateSubgridErrorShallowWater2D", 
    csubgridErrorCalculateSubgridErrorShallowWater2D, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the 2D shallow water equations"},
  { "calculateSubgridError_Harari_tau_sd", 
    csubgridErrorCalculateSubgridError_Harari_tau_sd, 
    METH_VARARGS, 
    "Calculate the subgrid error approximation for the scalar diffusion equation and time steps ala Harari formula scaled by dt"},
  { NULL,NULL,0,NULL}
};

PyMODINIT_FUNC initcsubgridError(void)
{
  PyObject *m,*d;
  m = Py_InitModule("csubgridError", csubgridErrorMethods);
  d = PyModule_GetDict(m);
  import_array();
}
/** @} */
