import sys
import setuptools
from distutils.core import setup, Extension
from Cython.Build import cythonize

import numpy
from Cython.Build import cythonize
from Cython.Distutils.extension import Extension
from Cython.Distutils import build_ext
## \file setup.py setup.py
#  \brief The python script for building proteus
#
#  Set the DISTUTILS_DEBUG environment variable to print detailed information while setup.py is running.
#

from proteus import config
from proteus.config import *

###to turn on debugging in c++
##\todo Finishing cleaning up setup.py/setup.cfg, config.py...
from distutils import sysconfig
cv = sysconfig.get_config_vars()
cv["CFLAGS"] = cv["CFLAGS"].replace("-DNDEBUG","")
cv["CFLAGS"] = cv["CFLAGS"].replace("-O3","")
cv["CFLAGS"] = cv["CFLAGS"].replace("-Wall","-w")
cv["CFLAGS"] = cv["CFLAGS"].replace("-Wstrict-prototypes","")


PROTEUS_PETSC_EXTRA_LINK_ARGS = getattr(config, 'PROTEUS_PETSC_EXTRA_LINK_ARGS', [])
PROTEUS_PETSC_EXTRA_COMPILE_ARGS = getattr(config, 'PROTEUS_PETSC_EXTRA_COMPILE_ARGS', [])
PROTEUS_CHRONO_CXX_FLAGS = getattr(config, 'PROTEUS_CHRONO_CXX_FLAGS', [])

proteus_install_path = os.path.join(sysconfig.get_python_lib(), 'proteus')

# handle non-system installations
for arg in sys.argv:
    if arg.startswith('--root'):
        proteus_install_path = proteus_install_path.partition(sys.prefix + '/')[-1]
        break
    if arg.startswith('--prefix'):
        proteus_install_path = proteus_install_path.partition(sys.prefix + '/')[-1]
        break

EXTENSIONS_TO_BUILD = [Extension("MeshAdaptPUMI.MeshAdaptPUMI",
                        sources = ['proteus/MeshAdaptPUMI/MeshAdaptPUMI.pyx', 'proteus/MeshAdaptPUMI/cMeshAdaptPUMI.cpp',
                                   'proteus/MeshAdaptPUMI/MeshConverter.cpp', 'proteus/MeshAdaptPUMI/ParallelMeshConverter.cpp',
                                   'proteus/MeshAdaptPUMI/MeshFields.cpp', 'proteus/MeshAdaptPUMI/SizeField.cpp',
                                   'proteus/MeshAdaptPUMI/DumpMesh.cpp',
                                   'proteus/MeshAdaptPUMI/ErrorResidualMethod.cpp','proteus/MeshAdaptPUMI/VMS.cpp'],
                        define_macros=[('PROTEUS_SUPERLU_H',PROTEUS_SUPERLU_H)],
                        language='c++',
                        include_dirs=[numpy.get_include(),'include',
                                      'proteus']+
                        PROTEUS_SCOREC_INCLUDE_DIRS,
                        library_dirs=PROTEUS_SCOREC_LIB_DIRS,
                        libraries=PROTEUS_SCOREC_LIBS,
                        extra_compile_args=PROTEUS_SCOREC_EXTRA_COMPILE_ARGS+PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_SCOREC_EXTRA_LINK_ARGS+PROTEUS_EXTRA_LINK_ARGS),
              Extension("mprans.cPres",['proteus/mprans/cPres.pyx'],
                        depends=['proteus/mprans/Pres.h', 'proteus/ModelFactory.h', 'proteus/CompKernel.h'],
                        language='c++',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus'],),
              Extension("mprans.cPresInit",['proteus/mprans/cPresInit.pyx'],
                        depends=['proteus/mprans/PresInit.h', 'proteus/ModelFactory.h', 'proteus/CompKernel.h'],
                        language='c++',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus']),
              Extension("mprans.cPresInc",['proteus/mprans/cPresInc.pyx'],
                        depends=['proteus/mprans/PresInc.h', 'proteus/ModelFactory.h', 'proteus/CompKernel.h'],
                        language='c++',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus']),
              Extension("mprans.cAddedMass",['proteus/mprans/cAddedMass.pyx'],
                        depends=['proteus/mprans/AddedMass.h', 'proteus/ModelFactory.h', 'proteus/CompKernel.h'],
                        language='c++',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus']),
              Extension("mprans.SedClosure",['proteus/mprans/SedClosure.pyx'],
                        depends=['proteus/mprans/SedClosure.h', 'proteus/ModelFactory.h', 'proteus/CompKernel.h'],
                        language='c++',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus']),
              Extension("mprans.cVOF3P",['proteus/mprans/cVOF3P.pyx'],
                        depends=['proteus/mprans/VOF3P.h', 'proteus/ModelFactory.h', 'proteus/CompKernel.h'],
                        language='c++',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus']),
              Extension("mprans.cVOS3P",['proteus/mprans/cVOS3P.pyx'],
                        depends=['proteus/mprans/VOS3P.h', 'proteus/ModelFactory.h', 'proteus/CompKernel.h'],
                        language='c++',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus']),
              Extension("mprans.cNCLS3P",['proteus/mprans/cNCLS3P.pyx'],
                        depends=['proteus/mprans/NCLS3P.h', 'proteus/ModelFactory.h', 'proteus/CompKernel.h'],
                        language='c++',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus']),
              Extension("mprans.cRDLS3P",['proteus/mprans/cRDLS3P.pyx'],
                        depends=['proteus/mprans/RDLS3P.h', 'proteus/ModelFactory.h', 'proteus/CompKernel.h'],
                        language='c++',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus']),
              Extension("mprans.cMCorr3P",
                        ["proteus/mprans/cMCorr3P.pyx"],
                        depends=["proteus/mprans/MCorr3P.h", 'proteus/ModelFactory.h', 'proteus/CompKernel.h'],
                        define_macros=[('PROTEUS_LAPACK_H',
                                        PROTEUS_LAPACK_H),
                                       ('PROTEUS_LAPACK_INTEGER',
                                        PROTEUS_LAPACK_INTEGER),
                                       ('PROTEUS_BLAS_H',
                                        PROTEUS_BLAS_H)],
                        language="c++",
                        include_dirs=[numpy.get_include(), 'proteus'],
                        library_dirs=[PROTEUS_LAPACK_LIB_DIR,
                                      PROTEUS_BLAS_LIB_DIR],
                        libraries=['m',PROTEUS_LAPACK_LIB,
                                   PROTEUS_BLAS_LIB],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension("richards.cRichards",['proteus/richards/cRichards.pyx'],
                        depends=['proteus/richards/Richards.h','proteus/ModelFactory.h', 'proteus/CompKernel.h'],
                        language='c++',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus']),
              Extension("elastoplastic.cElastoPlastic",
                        ['proteus/elastoplastic/cElastoPlastic.pyx'],
                        define_macros=[('PROTEUS_LAPACK_H',
                                        PROTEUS_LAPACK_H),
                                       ('PROTEUS_LAPACK_INTEGER',
                                        PROTEUS_LAPACK_INTEGER),
                                       ('PROTEUS_BLAS_H',
                                        PROTEUS_BLAS_H)],
                        depends=['proteus/elastoplastic/ElastoPlastic.h','proteus/ModelFactory.h', 'proteus/CompKernel.h'],
                        language='c++',
                        include_dirs=[numpy.get_include(),'proteus'],
                        library_dirs=[PROTEUS_LAPACK_LIB_DIR,
                                      PROTEUS_BLAS_LIB_DIR],
                        libraries=['m',PROTEUS_LAPACK_LIB,
                                   PROTEUS_BLAS_LIB],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension("mprans.cRANS3PF",['proteus/mprans/cRANS3PF.pyx'],
                        depends=['proteus/mprans/RANS3PF.h','proteus/mprans/RANS3PF2D.h', 'proteus/ModelFactory.h', 'proteus/CompKernel.h'],
                        language='c++',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus']),
              Extension("mprans.cRANS3PSed",['proteus/mprans/cRANS3PSed.pyx'],
                        depends=['proteus/mprans/RANS3PSed.h','proteus/mprans/RANS3PSed2D.h', 'proteus/ModelFactory.h', 'proteus/CompKernel.h'],
                        language='c++',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus']),
              Extension("Isosurface",['proteus/Isosurface.pyx'],
                        language='c',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus'],
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension("BoundaryConditions",['proteus/BoundaryConditions.py'],
                        language='c',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus']),
              Extension("mprans.BoundaryConditions",['proteus/mprans/BoundaryConditions.py'],
                        language='c',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus']),
              Extension("mprans.MeshSmoothing",['proteus/mprans/MeshSmoothing.pyx'],
                        language='c++',
                        include_dirs=[numpy.get_include(),'proteus',PROTEUS_INCLUDE_DIR],
                        libraries=['stdc++','m'],
                        extra_compile_args=["-std=c++11","-mavx"]),
              Extension("mprans.cMoveMeshMonitor",['proteus/mprans/cMoveMeshMonitor.pyx'],
                        language='c++',
                        include_dirs=[numpy.get_include(),'proteus',PROTEUS_INCLUDE_DIR],
                        libraries=['stdc++','m'],
                        extra_compile_args=["-std=c++11","-mavx"]),
              Extension("mbd.pyChronoCore",
                        sources=['proteus/mbd/pyChronoCore.pyx',
                                 'proteus/mbd/pyChronoCore.pxd',
                                 'proteus/mbd/ChVariablesBodyAddedMass.cpp',
                                 'proteus/mbd/ChBodyAddedMass.cpp',
                                 'proteus/mbd/ChronoHeaders.pxd'],
                        language='c++',
                        include_dirs=[numpy.get_include(),
                                      'proteus',
                                      PROTEUS_INCLUDE_DIR,
                                      PROTEUS_CHRONO_INCLUDE_DIR,
                                      PROTEUS_CHRONO_INCLUDE_DIR+'/chrono',
                                      PROTEUS_CHRONO_INCLUDE_DIR+'/chrono/collision/bullet',],
                        library_dirs=[PROTEUS_CHRONO_LIB_DIR],
                        libraries=['ChronoEngine',
                                   'ChronoEngine_fea',
                                   'stdc++',
                                   'm'],
                        extra_compile_args=PROTEUS_CHRONO_CXX_FLAGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension("mbd.ChRigidBody",
                        sources=['proteus/mbd/ChRigidBody.pyx'],
                        depends=['proteus/mbd/ChRigidBody.h',
                                 'proteus/mbd/ChMoorings.h'],
                        language='c++',
                        include_dirs=[numpy.get_include(),
                                      'proteus',
                                      PROTEUS_INCLUDE_DIR,
                                      PROTEUS_CHRONO_INCLUDE_DIR,
                                      PROTEUS_CHRONO_INCLUDE_DIR+'/chrono',
                                      PROTEUS_CHRONO_INCLUDE_DIR+'/chrono/collision/bullet',],
                        library_dirs=[PROTEUS_CHRONO_LIB_DIR],
                        libraries=['ChronoEngine',
                                   'ChronoEngine_fea',
                                   'stdc++',
                                   'm'],
                        extra_compile_args=PROTEUS_CHRONO_CXX_FLAGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension("WaveTools",['proteus/WaveTools.py'],
                        depends=['proteus/WaveTools.h'],
                        language='c++',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus']),
              Extension("fenton.Fenton",
                         sources=['proteus/fenton/Fenton.pyx',
                                  'proteus/fenton/Solve.cpp',
                                  'proteus/fenton/Dpythag.cpp',
                                  'proteus/fenton/Dsvbksb.cpp',
                                  'proteus/fenton/Dsvdcmp.cpp',
                                  'proteus/fenton/Inout.cpp',
                                  'proteus/fenton/Subroutines.cpp',
                                  'proteus/fenton/Util.cpp',],
                         language='c++',
                         include_dirs=[numpy.get_include(),
                                       'proteus',
                                       PROTEUS_INCLUDE_DIR,
                                       PROTEUS_NCURSES_INCLUDE_DIR,],
                         library_dirs=[PROTEUS_NCURSES_LIB_DIR,],
                         libraries=['ncurses','stdc++','m'],
                         extra_compile_args=["-std=c++11"]),
              Extension("ADR",['proteus/ADR.pyx'],
                        depends=['proteus/ADR.h', 'proteus/ModelFactory.h', 'proteus/CompKernel.h'],
                        language='c++',
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(),'proteus']),
              Extension("waveFunctions",['proteus/waveFunctions.pyx','proteus/transportCoefficients.c'],
                        include_dirs=[numpy.get_include(),'proteus'],
                        extra_compile_args=PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension("subsurfaceTransportFunctions",['proteus/subsurfaceTransportFunctions.pyx'],
                        include_dirs=[numpy.get_include(),'proteus'],
                        extra_compile_args=PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension('cfemIntegrals',
                        ['proteus/cfemIntegralsModule.c','proteus/femIntegrals.c','proteus/postprocessing.c'],
                        define_macros=[('PROTEUS_SUPERLU_H',PROTEUS_SUPERLU_H),
                                       ('PROTEUS_LAPACK_H',PROTEUS_LAPACK_H),
                                       ('PROTEUS_LAPACK_INTEGER',PROTEUS_LAPACK_INTEGER),
                                       ('PROTEUS_BLAS_H',PROTEUS_BLAS_H)],
                        include_dirs=[numpy.get_include(),'proteus',
                                      PROTEUS_SUPERLU_INCLUDE_DIR,
                                      PROTEUS_LAPACK_INCLUDE_DIR,
                                      PROTEUS_BLAS_INCLUDE_DIR],
                        library_dirs=[PROTEUS_LAPACK_LIB_DIR,
                                      PROTEUS_BLAS_LIB_DIR],
                        libraries=['m',PROTEUS_LAPACK_LIB,PROTEUS_BLAS_LIB],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension('cmeshTools',
                        ['proteus/cmeshToolsModule.cpp','proteus/mesh.cpp','proteus/meshio.cpp'],
                        define_macros=[('PROTEUS_SUPERLU_H',PROTEUS_SUPERLU_H),
                                       ('PROTEUS_TRIANGLE_H',PROTEUS_TRIANGLE_H)],
                        include_dirs=([numpy.get_include(),'proteus']+
                                      [PROTEUS_TRIANGLE_INCLUDE_DIR]),
                        libraries=['m',PROTEUS_DAETK_LIB]+[PROTEUS_TRIANGLE_LIB],
                        library_dirs=[PROTEUS_DAETK_LIB_DIR]+[PROTEUS_TRIANGLE_LIB_DIR],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension('ctransportCoefficients',
                        ['proteus/ctransportCoefficientsModule.c','proteus/transportCoefficients.c'],
                        include_dirs=[numpy.get_include(),'proteus'],
                        libraries=['m'],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension('csubgridError',
                        ['proteus/csubgridErrorModule.c','proteus/subgridError.c'],
                        include_dirs=[numpy.get_include(),'proteus'],
                        libraries=['m'],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension('cshockCapturing',
                        ['proteus/cshockCapturingModule.c','proteus/shockCapturing.c'],
                        include_dirs=[numpy.get_include(),'proteus'],
                        libraries=['m'],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension('lapackWrappers',
                        ['proteus/lapackWrappersModule.c'],
                        define_macros=[('PROTEUS_LAPACK_H',PROTEUS_LAPACK_H),
                                       ('PROTEUS_LAPACK_INTEGER',PROTEUS_LAPACK_INTEGER),
                                       ('PROTEUS_BLAS_H',PROTEUS_BLAS_H)],
                        include_dirs=[numpy.get_include(),'proteus',
                                      PROTEUS_LAPACK_INCLUDE_DIR,
                                      PROTEUS_BLAS_INCLUDE_DIR],
                        library_dirs=[PROTEUS_LAPACK_LIB_DIR,PROTEUS_BLAS_LIB_DIR],
                        libraries=['m',
                                   PROTEUS_LAPACK_LIB,
                                   PROTEUS_BLAS_LIB],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension('superluWrappers',
                        ['proteus/superluWrappersModule.c'],
                        define_macros=[('PROTEUS_SUPERLU_H',PROTEUS_SUPERLU_H),
                                       ('PROTEUS_BLAS_H',PROTEUS_BLAS_H)],
                        include_dirs=[numpy.get_include(),'proteus',PROTEUS_SUPERLU_INCLUDE_DIR],
                        library_dirs=[PROTEUS_SUPERLU_LIB_DIR,PROTEUS_LAPACK_LIB_DIR,PROTEUS_BLAS_LIB_DIR],
                        libraries=['m',PROTEUS_SUPERLU_LIB,PROTEUS_LAPACK_LIB,PROTEUS_BLAS_LIB],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension('triangleWrappers',
                        ['proteus/triangleWrappersModule.c'],
                        define_macros=[('PROTEUS_TRIANGLE_H',
                                        PROTEUS_TRIANGLE_H),
                                       ('MWF_ADDED_FLAGS',
                                        1)],
                        include_dirs=[numpy.get_include(),PROTEUS_TRIANGLE_INCLUDE_DIR],
                        library_dirs=[PROTEUS_TRIANGLE_LIB_DIR],
                        libraries=['m',
                                   PROTEUS_TRIANGLE_LIB],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension('csmoothers',
                        ['proteus/csmoothersModule.c', 'proteus/smoothers.c'],
                        define_macros=[('PROTEUS_SUPERLU_H',PROTEUS_SUPERLU_H),
                                       ('PROTEUS_LAPACK_H',PROTEUS_LAPACK_H),
                                       ('PROTEUS_LAPACK_INTEGER',PROTEUS_LAPACK_INTEGER),
                                       ('PROTEUS_BLAS_H',PROTEUS_BLAS_H)],
                        include_dirs=[numpy.get_include(),'proteus',
                                      PROTEUS_SUPERLU_INCLUDE_DIR,
                                      PROTEUS_LAPACK_INCLUDE_DIR,
                                      PROTEUS_BLAS_INCLUDE_DIR
                        ],
                        library_dirs=[PROTEUS_SUPERLU_INCLUDE_DIR,
                                      PROTEUS_SUPERLU_LIB_DIR,
                                      PROTEUS_LAPACK_LIB_DIR,
                                      PROTEUS_BLAS_LIB_DIR],
                        libraries=['m',
                                   PROTEUS_SUPERLU_LIB,
                                   PROTEUS_LAPACK_LIB,
                                   PROTEUS_BLAS_LIB],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension('ctimeIntegration',
                        ['proteus/ctimeIntegrationModule.c','proteus/timeIntegration.c'],
                        include_dirs=[numpy.get_include(),'proteus'],
                        libraries=['m'],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension("canalyticalSolutions",["proteus/canalyticalSolutions.pyx", "proteus/analyticalSolutions.c"],
                        depends=["proteus/analyticalSolutions.h"],
                        extra_compile_args=PROTEUS_OPT,
                        language="c", include_dirs=[numpy.get_include(), 'proteus']),
              Extension('cpostprocessing',
                        ['proteus/cpostprocessingModule.c','proteus/postprocessing.c','proteus/femIntegrals.c'],
                        define_macros=[('MWF_ADDED_FLAGS',
                                        1),
                                       ('PROTEUS_LAPACK_H',PROTEUS_LAPACK_H),
                                       ('PROTEUS_LAPACK_INTEGER',PROTEUS_LAPACK_INTEGER)
                        ],
                        include_dirs=[numpy.get_include(),'proteus',
                                      PROTEUS_LAPACK_INCLUDE_DIR
                        ],
                        library_dirs=[PROTEUS_LAPACK_LIB_DIR,PROTEUS_BLAS_LIB_DIR],
                        libraries=['m',PROTEUS_LAPACK_LIB,PROTEUS_BLAS_LIB],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension('cnumericalFlux',
                        ['proteus/cnumericalFluxModule.c','proteus/numericalFlux.c'],
                        include_dirs=[numpy.get_include(),'proteus'],
                        libraries=['m'],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension('cTwophaseDarcyCoefficients',
                        ['proteus/cTwophaseDarcyCoefficientsModule.cpp','proteus/SubsurfaceTransportCoefficients.cpp'],
                        include_dirs=[numpy.get_include(),'proteus'],
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS,
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT),
              Extension('cpskRelations',
                        ['proteus/cpskRelationsModule.cpp','proteus/SubsurfaceTransportCoefficients.cpp'],
                        include_dirs=['proteus'],
                        libraries=['m'],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension('cSubsurfaceTransportCoefficients',
                        ['proteus/cSubsurfaceTransportCoefficientsModule.cpp','proteus/SubsurfaceTransportCoefficients.cpp'],
                        include_dirs=[numpy.get_include(),'proteus'],
                        libraries=['m'],
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS,
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT),
              Extension('flcbdfWrappers',
                        ['proteus/flcbdfWrappersModule.cpp','proteus/mesh.cpp','proteus/meshio.cpp'],
                        define_macros=[('PROTEUS_TRIANGLE_H',PROTEUS_TRIANGLE_H),
                                       ('PROTEUS_SUPERLU_H',PROTEUS_SUPERLU_H),
                                       ('CMRVEC_BOUNDS_CHECK',1),
                                       ('MV_VECTOR_BOUNDS_CHECK',1),
                                       ('PETSCVEC_BOUNDS_CHECK',1),
                                       ('F77_POST_UNDERSCORE',1),
                                       ('USE_BLAS',1)],
                        include_dirs=['proteus',
                                      numpy.get_include(),
                                      PROTEUS_SUPERLU_INCLUDE_DIR,
                                      PROTEUS_TRIANGLE_INCLUDE_DIR,
                                      PROTEUS_DAETK_INCLUDE_DIR,
                                      PROTEUS_HDF5_INCLUDE_DIR] + \
                        PROTEUS_PETSC_INCLUDE_DIRS + \
                        PROTEUS_MPI_INCLUDE_DIRS,
                        library_dirs=[PROTEUS_DAETK_LIB_DIR]+PROTEUS_PETSC_LIB_DIRS+PROTEUS_MPI_LIB_DIRS+PROTEUS_HDF5_LIB_DIRS,
                        libraries=['hdf5','stdc++','m',PROTEUS_DAETK_LIB]+PROTEUS_PETSC_LIBS+PROTEUS_MPI_LIBS+PROTEUS_HDF5_LIBS,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS + PROTEUS_PETSC_EXTRA_LINK_ARGS,
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS + PROTEUS_PETSC_EXTRA_COMPILE_ARGS+PROTEUS_OPT),
              Extension("mprans.cCLSVOF",["proteus/mprans/cCLSVOF.pyx"],
                        depends=["proteus/mprans/CLSVOF.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        language="c++",
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(), 'proteus']),
              Extension("mprans.cNCLS",["proteus/mprans/cNCLS.pyx"],
                        depends=["proteus/mprans/NCLS.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        language="c++",
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(), 'proteus']),
              Extension("mprans.cMCorr",["proteus/mprans/cMCorr.pyx"],
                        depends=["proteus/mprans/MCorr.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        define_macros=[('PROTEUS_LAPACK_H',PROTEUS_LAPACK_H),
                                       ('PROTEUS_LAPACK_INTEGER',PROTEUS_LAPACK_INTEGER),
                                       ('PROTEUS_BLAS_H',PROTEUS_BLAS_H)],language="c++",
                        include_dirs=[numpy.get_include(), 'proteus'],
                        library_dirs=[PROTEUS_LAPACK_LIB_DIR,
                                      PROTEUS_BLAS_LIB_DIR],
                        libraries=['m',PROTEUS_LAPACK_LIB,PROTEUS_BLAS_LIB],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS),
              Extension("mprans.cRANS2P",["proteus/mprans/cRANS2P.pyx"],
                        depends=["proteus/mprans/RANS2P.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        extra_compile_args=PROTEUS_OPT,
                        language="c++", include_dirs=[numpy.get_include(), 'proteus']),
              Extension("mprans.cRANS2P2D",["proteus/mprans/cRANS2P2D.pyx"],
                        depends=["proteus/mprans/RANS2P2D.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        language="c++",
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(), 'proteus']),
              Extension("mprans.cRDLS",["proteus/mprans/cRDLS.pyx"],
                        depends=["proteus/mprans/RDLS.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        language="c++",
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(), 'proteus']),
              Extension("mprans.cVOF",["proteus/mprans/cVOF.pyx"],
                        depends=["proteus/mprans/VOF.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        language="c++",
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(), 'proteus']),
              Extension("mprans.cBlendedSpaces",["proteus/mprans/cBlendedSpaces.pyx"],
                        depends=["proteus/mprans/BlendedSpaces.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        language="c++",
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(), 'proteus']),
              Extension("mprans.cMoveMesh",["proteus/mprans/cMoveMesh.pyx"],
                        depends=["proteus/mprans/MoveMesh.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        language="c++",
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(), 'proteus']),
              Extension("mprans.cMoveMesh2D",["proteus/mprans/cMoveMesh2D.pyx"],
                        depends=["proteus/mprans/MoveMesh2D.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        language="c++",
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(), 'proteus']),
              Extension("mprans.cSW2D",["proteus/mprans/cSW2D.pyx"],
                        depends=["proteus/mprans/SW2D.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        language="c++",
                        include_dirs=[numpy.get_include(), 'proteus'],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+['-g']+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS+['-g']),
              Extension("mprans.cSW2DCV",["proteus/mprans/cSW2DCV.pyx"],
                        depends=["proteus/mprans/SW2DCV.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        language="c++",
                        include_dirs=[numpy.get_include(), 'proteus'],
                        extra_compile_args=PROTEUS_EXTRA_COMPILE_ARGS+['-g']+PROTEUS_OPT,
                        extra_link_args=PROTEUS_EXTRA_LINK_ARGS+['-g']),
              Extension("mprans.cKappa",["proteus/mprans/cKappa.pyx"],
                        depends=["proteus/mprans/Kappa.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        language="c++",
                        include_dirs=[numpy.get_include(), 'proteus']),
              Extension("mprans.cKappa2D",["proteus/mprans/cKappa2D.pyx"],
                        depends=["proteus/mprans/Kappa2D.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        language="c++",
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(), 'proteus']),
              Extension("mprans.cDissipation",["proteus/mprans/cDissipation.pyx"],
                        depends=["proteus/mprans/Dissipation.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        language="c++",
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(), 'proteus']),
              Extension("mprans.cDissipation2D",["proteus/mprans/cDissipation2D.pyx"],
                        depends=["proteus/mprans/Dissipation2D.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        language="c++",
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(), 'proteus']),
              Extension("mprans.cRANS2P_IB",["proteus/mprans/cRANS2P_IB.pyx"],
                        depends=["proteus/mprans/RANS2P_IB.h"] + ["proteus/ModelFactory.h","proteus/CompKernel.h"],
                        language="c++",
                        extra_compile_args=PROTEUS_OPT,
                        include_dirs=[numpy.get_include(), 'proteus'])
]

def setup_given_extensions(extensions):
    setup(name='proteus',
          version='1.5.1.hpFEM',
          classifiers=[
              'Development Status :: 4 - Beta',
              'Environment :: Console',
              'Environment :: Web Environment',
              'Intended Audience :: End Users/Desktop',
              'Intended Audience :: Developers',
              'Intended Audience :: Science/Research',
              'License :: OSI Approved :: MIT License',
              'Programming Language :: Python :: 2.7'
              'Programming Language :: Python :: 2 :: Only',
              'Programming Language :: Python :: Implementation :: CPython',
              'Topic :: Scientific/Engineering :: Mathematics',
              'Topic :: Scientific/Engineering :: Physics',
              'Operating System :: MacOS :: MacOS X',
              'Operating System :: POSIX',
          ],
          description='Python tools for multiphysics modeling',
          author='The Proteus Developers',
          author_email='proteus-dev@googlegroups.com',
          url='http://proteustoolkit.org',
          packages = ['proteus',
                      'proteus.fenton',
                      'proteus.mprans',
                      'proteus.richards',
                      'proteus.elastoplastic',
                      'proteus.mbd',
                      'proteus.test_utils',
                      'proteus.config',
                      'proteus.tests',
                      'proteus.tests.ci',
                      'proteus.tests.griffiths_lane_6',
                      'proteus.tests.levelset',
                      'proteus.tests.linalgebra_tests',
                      'proteus.tests.LS_with_edgeBased_EV',
                      'proteus.tests.LS_with_edgeBased_EV.VOF',
                      'proteus.tests.LS_with_edgeBased_EV.NCLS',
                      'proteus.tests.BernsteinPolynomials',
                      'proteus.tests.BernsteinPolynomials.poisson_eqn',
                      'proteus.tests.elliptic_redist',
                      'proteus.tests.elliptic_redist.RDLS',
                      'proteus.tests.elliptic_redist.RDLS3P',
                      'proteus.tests.surface_tension',
                      'proteus.tests.surface_tension.rising_bubble_rans3p',
                      'proteus.tests.CLSVOF',
                      'proteus.tests.CLSVOF.disc_ICs',
                      'proteus.tests.CLSVOF.with_RANS2P',
                      'proteus.tests.CLSVOF.with_RANS3PF',
                      'proteus.tests.CLSVOF.pure_level_set',
                      'proteus.tests.SWEs',
                      'proteus.tests.SWEs.dam_over_bumps',
                      'proteus.tests.SWEs.oneD_dambreak_flat_bottom',
                      'proteus.tests.SWEs.paraboloid_with_friction',
                      'proteus.tests.SWEs.paraboloid_with_friction.oneD',
                      'proteus.tests.SWEs.paraboloid_with_friction.twoD',
                      'proteus.tests.SWEs.test_gauges',
                      'proteus.tests.SWEs.test_reflecting_BCs',
                      'proteus.tests.matrix_constructor',
                      'proteus.tests.matrix_constructor.import_modules',
                      'proteus.MeshAdaptPUMI',
                      'proteus.tests.MeshAdaptPUMI',
                      'proteus.tests.MeshAdaptPUMI.gauge_compare.dambreak_Colagrossi_2D',
                      'proteus.tests.mesh_tests',
                      'proteus.tests.mesh_tests.import_modules',
                      'proteus.tests.periodic',
                      'proteus.tests.periodic.petsc',
                      'proteus.tests.periodic.comparison_files',
                      'proteus.tests.poisson_2d',
                      'proteus.tests.post_processing',
                      'proteus.tests.post_processing.import_modules',
                      'proteus.tests.ProjScheme_with_EV',
                      'proteus.tests.single_phase_gw',
                      'proteus.tests.solver_tests',
                      'proteus.tests.solver_tests.import_modules',
                      'proteus.tests.solver_tests_slow',
                      'proteus.tests.solver_tests_slow.import_modules',
                      'proteus.tests.solver_tests_mprans',
                      'proteus.tests.solver_tests_mprans.import_modules',
                      'proteus.tests.cylinder2D',
                      'proteus.tests.cylinder2D.conforming_rans2p',
                      'proteus.tests.cylinder2D.conforming_rans3p',
                      'proteus.tests.cylinder2D.ibm_method',
                      'proteus.tests.cylinder2D.ibm_rans2p',
                      'proteus.tests.cylinder2D.ibm_rans2p_3D',
                      'proteus.tests.cylinder2D.sbm_method',
                      'proteus.tests.cylinder2D.sbm_3Dmesh',
                      'proteus.tests.HotStart_3P',
                      'proteus.tests.AddedMass',
                      'proteus.tests.MoveMeshMonitor',
                      'proteus.tests.wave_tests',
          ],
          cmdclass = {'build_ext':build_ext},
          ext_package='proteus',
          ext_modules=extensions,
          data_files=[(proteus_install_path,
                       ['proteus/proteus_blas.h',
                        'proteus/proteus_lapack.h',
                        'proteus/ModelFactory.h',
                        'proteus/CompKernel.h'
                       ]),
                      (os.path.join(proteus_install_path,'tests'),
                       ['proteus/tests/hex_cube_3x3.xmf',
                        'proteus/tests/hex_cube_3x3.h5']),
                      (os.path.join(proteus_install_path,'tests','linalgebra_tests'),
                       ['proteus/tests/linalgebra_tests/sparse_mat_1.txt',
                        'proteus/tests/linalgebra_tests/jac.bin']),
                      (os.path.join(proteus_install_path,'tests','griffiths_lane_6'),
                       ['proteus/tests/griffiths_lane_6/richards_expected.h5',
                        'proteus/tests/griffiths_lane_6/elastoplastic_expected.h5']),
                      (os.path.join(proteus_install_path,'tests','levelset'),
                       ['proteus/tests/levelset/rotation/rotation_c0p1cg_vbdf_2_level_1_expected.h5',
                        'proteus/tests/levelset/vortex2D/vortex_c0p1cg_vbdf_2_level_1_expected.h5',
                        'proteus/tests/levelset/vortex/vortex_c0p1cg_bdf_2_level_1_expected.h5']),
                      (os.path.join(proteus_install_path,'tests','ci','comparison_files'),
                       ['proteus/tests/ci/comparison_files/floating_bar.h5',
                        'proteus/tests/ci/comparison_files/phi_t_0.000000_000.tgz']),
                      (os.path.join(proteus_install_path,'tests','LS_with_edgeBased_EV','VOF','comparison_files'),
                       ['proteus/tests/LS_with_edgeBased_EV/VOF/comparison_files/vof_level_3_SmoothnessBased.h5',
                        'proteus/tests/LS_with_edgeBased_EV/VOF/comparison_files/vof_level_3_EV1.h5',
                        'proteus/tests/LS_with_edgeBased_EV/VOF/comparison_files/vof_level_3_SUPG.h5',
                        'proteus/tests/LS_with_edgeBased_EV/VOF/comparison_files/vof_level_3_EV2.h5']),
                      (os.path.join(proteus_install_path,'tests','LS_with_edgeBased_EV','NCLS','comparison_files'),
                       ['proteus/tests/LS_with_edgeBased_EV/NCLS/comparison_files/ncls_level_3_non_saturated_ls.h5',
                        'proteus/tests/LS_with_edgeBased_EV/NCLS/comparison_files/ncls_level_3_pureAdvection_SUPG.h5',
                        'proteus/tests/LS_with_edgeBased_EV/NCLS/comparison_files/ncls_level_3_pureAdvection_EV1.h5',
                        'proteus/tests/LS_with_edgeBased_EV/NCLS/comparison_files/ncls_level_3_saturated_ls.h5']),
                      (os.path.join(proteus_install_path,'tests','BernsteinPolynomials','poisson_eqn','comparison_files'),
                       ['proteus/tests/BernsteinPolynomials/poisson_eqn/comparison_files/2D_poisson_hex_degree2.h5',
                        'proteus/tests/BernsteinPolynomials/poisson_eqn/comparison_files/2D_poisson_simplex_degree2.h5',
                        'proteus/tests/BernsteinPolynomials/poisson_eqn/comparison_files/3D_poisson_hex_degree2.h5',
                        'proteus/tests/BernsteinPolynomials/poisson_eqn/comparison_files/3D_poisson_simplex_degree2.h5']),
                      (os.path.join(proteus_install_path,'tests','surface_tension','rising_bubble_rans3p','comparison_files'),
                       ['proteus/tests/surface_tension/rising_bubble_rans3p/comparison_files/risingBubble_2D_supg.h5',
                        'proteus/tests/surface_tension/rising_bubble_rans3p/comparison_files/risingBubble_2D_ev.h5',
                        'proteus/tests/surface_tension/rising_bubble_rans3p/comparison_files/risingBubble_3D_supg.h5',
                        'proteus/tests/surface_tension/rising_bubble_rans3p/comparison_files/risingBubble_3D_ev.h5']),                      
                      (os.path.join(proteus_install_path,'tests','CLSVOF','disc_ICs','comparison_files'),
                       ['proteus/tests/CLSVOF/disc_ICs/comparison_files/test_case_1.h5',
                        'proteus/tests/CLSVOF/disc_ICs/comparison_files/test_case_2.h5']),                      
                      (os.path.join(proteus_install_path,'tests','CLSVOF','pure_level_set','comparison_files'),
                       ['proteus/tests/CLSVOF/pure_level_set/comparison_files/clsvof_test_case_1.h5',
                        'proteus/tests/CLSVOF/pure_level_set/comparison_files/clsvof_test_case_2.h5',
                        'proteus/tests/CLSVOF/pure_level_set/comparison_files/clsvof_test_case_3.h5',
                        'proteus/tests/CLSVOF/pure_level_set/comparison_files/clsvof_test_case_4.h5']),                      
                      (os.path.join(proteus_install_path,'tests','CLSVOF','with_RANS2P','comparison_files'),
                       ['proteus/tests/CLSVOF/with_RANS2P/comparison_files/multiphase_2D_falling_bubble.h5']),
                      (os.path.join(proteus_install_path,'tests','CLSVOF','with_RANS3PF','comparison_files'),
                       ['proteus/tests/CLSVOF/with_RANS3PF/comparison_files/multiphase_2D_falling_bubble.h5',
                        'proteus/tests/CLSVOF/with_RANS3PF/comparison_files/multiphase_3D_falling_bubble.h5']),
                      (os.path.join(proteus_install_path,'tests','SWEs','dam_over_bumps','comparison_files'),
                       ['proteus/tests/SWEs/dam_over_bumps/comparison_files/SWEs_dam_over_bumps.h5']),
                      (os.path.join(proteus_install_path,'tests','SWEs','oneD_dambreak_flat_bottom','comparison_files'),
                       ['proteus/tests/SWEs/oneD_dambreak_flat_bottom/comparison_files/SWEs_oneD_dambreak_flat_bottom.h5']),
                      (os.path.join(proteus_install_path,'tests','SWEs','paraboloid_with_friction','oneD','comparison_files'),
                       ['proteus/tests/SWEs/paraboloid_with_friction/oneD/comparison_files/SWEs_oneD_paraboloid_with_friction.h5']),
                      (os.path.join(proteus_install_path,'tests','SWEs','paraboloid_with_friction','twoD','comparison_files'),
                       ['proteus/tests/SWEs/paraboloid_with_friction/twoD/comparison_files/SWEs_twoD_paraboloid_with_friction.h5']),
                      (os.path.join(proteus_install_path,'tests','SWEs','test_gauges','comparison_files'),
                       ['proteus/tests/SWEs/test_gauges/comparison_files/SWEs_test_gauges.h5']),
                      (os.path.join(proteus_install_path,'tests','SWEs','test_reflecting_BCs','comparison_files'),
                       ['proteus/tests/SWEs/test_reflecting_BCs/comparison_files/SWEs_test_reflecting_BCs.h5']),
                      (os.path.join(proteus_install_path,'tests','solver_tests','import_modules'),
                       ['proteus/tests/solver_tests/import_modules/quad_mass_matrix.npy',
                        'proteus/tests/solver_tests/import_modules/sol_10.npy',
                        'proteus/tests/solver_tests/import_modules/sol_20_lst.npy',
                        'proteus/tests/solver_tests/import_modules/input_vec_tppcd.bin',
                        'proteus/tests/solver_tests/import_modules/tp_pcd_y_output.bin',
                        'proteus/tests/solver_tests/import_modules/tppcd_y_dirichlet_dof.bin',
                        'proteus/tests/solver_tests/import_modules/Qp_visc.bin',
                        'proteus/tests/solver_tests/import_modules/Qp_dens.bin',
                        'proteus/tests/solver_tests/import_modules/Ap_rho.bin',
                        'proteus/tests/solver_tests/import_modules/Np_rho.bin',
                        'proteus/tests/solver_tests/import_modules/saddle_point_small.bin',
                        'proteus/tests/solver_tests/import_modules/saddle_point_matrix.bin',
                        'proteus/tests/solver_tests/import_modules/rans2p_step_newton_1.bin',
                        'proteus/tests/solver_tests/import_modules/rans2p_step_newton_5.bin',
                        'proteus/tests/solver_tests/import_modules/NSE_cavity_matrix.bin',
                        'proteus/tests/solver_tests/import_modules/NSE_step_no_slip.bin']),
                      (os.path.join(proteus_install_path,'tests','mesh_tests','comparison_files'),
                       ['proteus/tests/mesh_tests/comparison_files/poiseulle_xmf.output',
                        'proteus/tests/mesh_tests/comparison_files/poiseulle_global_xmf.output']),
                      (os.path.join(proteus_install_path,'tests','solver_tests_slow','comparison_files'),
                       ['proteus/tests/solver_tests_slow/comparison_files/Qp_expected.log',
                        'proteus/tests/solver_tests_slow/comparison_files/drivenCavityStokes_expected.h5']),
                      (os.path.join(proteus_install_path,'tests','matrix_constructor','comparison_files'),
                       ['proteus/tests/matrix_constructor/comparison_files/mass_reference_c0p1_2D.txt',
                        'proteus/tests/matrix_constructor/comparison_files/mass_reference_TH_2D.npy']),
                      (os.path.join(proteus_install_path,'tests','periodic','petsc'),
                       ['proteus/tests/periodic/petsc/petsc.options.schur.selfp_petsc.amg',
                        'proteus/tests/periodic/petsc/petsc.options.schur.selfp_petsc.gamg.superlu',
                        'proteus/tests/periodic/petsc/petsc.options.schur.selfp_petsc.superlu']),
                      (os.path.join(proteus_install_path,'tests','periodic','comparison_files'),
                       ['proteus/tests/periodic/comparison_files/basic_2d_test.h5',
                        'proteus/tests/periodic/comparison_files/basic_3d_test.h5']),
                      (os.path.join(proteus_install_path,'tests','post_processing','import_modules'),
                       ['proteus/tests/post_processing/import_modules/reference_simplex_keep.ele',
                        'proteus/tests/post_processing/import_modules/reference_simplex_keep.face',
                        'proteus/tests/post_processing/import_modules/reference_simplex_keep.node',
                        'proteus/tests/post_processing/import_modules/reference_simplex_keep.poly',
                        'proteus/tests/post_processing/import_modules/bdm2_3d_face_func_vals.data',
                        'proteus/tests/post_processing/import_modules/bdm2_3d_interior_func_vals.data',
                        'proteus/tests/post_processing/import_modules/bdm_bdy_func_values_3dmesh.data',
                        'proteus/tests/post_processing/import_modules/bdm_func_values_3dmesh.data']),
                      (os.path.join(proteus_install_path,'tests','post_processing','comparison_files'),
                       ['proteus/tests/post_processing/comparison_files/BDM_Test_File.h5',
                        'proteus/tests/post_processing/comparison_files/bdm2_ref_proj_mat.txt',
                        'proteus/tests/post_processing/comparison_files/bdm2_reference_simplex_mat.data',
                        'proteus/tests/post_processing/comparison_files/bdm2_reference_simplex_rhs.data',
                        'proteus/tests/post_processing/comparison_files/bdm_bdy_func_values.npy',
                        'proteus/tests/post_processing/comparison_files/bdm_bdy_func_values_mesh_8.npy',
                        'proteus/tests/post_processing/comparison_files/bdm_bdy_func_values_trig.npy',
                        'proteus/tests/post_processing/comparison_files/bdm_func_values.npy',
                        'proteus/tests/post_processing/comparison_files/bdm_func_values_mesh_8.npy',
                        'proteus/tests/post_processing/comparison_files/bdm_func_values_trig.npy',
                        'proteus/tests/post_processing/comparison_files/poisson_bdm1_test.h5',
                        'proteus/tests/post_processing/comparison_files/test_bdm2_sshaped_region_expected.h5',
                        'proteus/tests/post_processing/comparison_files/test_bdm_sshaped_region_expected.h5',
                        'proteus/tests/post_processing/comparison_files/trig_velocity_rep.npy']),
                      (os.path.join(proteus_install_path,'tests','matrix_constructor','comparison_files'),
                       ['proteus/tests/matrix_constructor/comparison_files/velocity_laplace_C0P2_mesh.npy',
                        'proteus/tests/matrix_constructor/comparison_files/single_phase_THQuad_4_expected.data']),
                      (os.path.join(proteus_install_path,'tests','solver_tests_slow','comparison_files'),
                       ['proteus/tests/solver_tests_slow/comparison_files/drivenCavityNSE_LSC_expected.h5',
                        'proteus/tests/solver_tests_slow/comparison_files/drivenCavityNSE_LSC_expected.xmf',
                        'proteus/tests/solver_tests_slow/comparison_files/drivenCavityNSE_LSC_expected.log']),
                      (os.path.join(proteus_install_path,'tests','solver_tests_slow','import_modules'),
                       ['proteus/tests/solver_tests_slow/import_modules/petsc.options.schur_lsc']),
                      (os.path.join(proteus_install_path,'tests','solver_tests_mprans','comparison_files'),
                       ['proteus/tests/solver_tests_mprans/comparison_files/twp_navier_stokes_cavity_2d.h5',
                        'proteus/tests/solver_tests_mprans/comparison_files/twp_navier_stokes_cavity_2d.xmf']),
                      (os.path.join(proteus_install_path,'tests','MeshAdaptPUMI'),
                       ['proteus/tests/MeshAdaptPUMI/cube0.smb',
                        'proteus/tests/MeshAdaptPUMI/cube.dmg',
                        'proteus/tests/MeshAdaptPUMI/Couette.null',
                        'proteus/tests/MeshAdaptPUMI/Couette.msh',
                        'proteus/tests/MeshAdaptPUMI/Couette2D.msh',
                        'proteus/tests/MeshAdaptPUMI/Rectangle0.smb',
                        'proteus/tests/MeshAdaptPUMI/Rectangle1.smb',
                        'proteus/tests/MeshAdaptPUMI/Rectangle.dmg',
                        'proteus/tests/MeshAdaptPUMI/TwoQuads0.smb',
                        'proteus/tests/MeshAdaptPUMI/TwoQuads.dmg']),
                      (os.path.join(proteus_install_path,'tests','MeshAdaptPUMI','gauge_compare','dambreak_Colagrossi_2D'),
                       ['proteus/tests/MeshAdaptPUMI/gauge_compare/dambreak_Colagrossi_2D/Reconstructed.dmg',
                        'proteus/tests/MeshAdaptPUMI/gauge_compare/dambreak_Colagrossi_2D/Reconstructed0.smb']),
                      (os.path.join(proteus_install_path,'tests','poisson_2d'),
                       ['proteus/tests/poisson_2d/square4x4.3dm',
                        'proteus/tests/poisson_2d/square4x4.bc']),
                      (os.path.join(proteus_install_path,'tests','cylinder2D','conforming_rans3p','comparison_files'),
                       ['proteus/tests/cylinder2D/conforming_rans3p/comparison_files/T1P1.h5',
                        'proteus/tests/cylinder2D/conforming_rans3p/comparison_files/T4P2.h5',
                        'proteus/tests/cylinder2D/conforming_rans3p/comparison_files/T8P2.h5']),
                      (os.path.join(proteus_install_path,'tests','cylinder2D','ibm_method','comparison_files'),
                       ['proteus/tests/cylinder2D/ibm_method/comparison_files/T1_rans3p.h5']),
                      (os.path.join(proteus_install_path,'tests','cylinder2D','ibm_rans2p','comparison_files'),
                       ['proteus/tests/cylinder2D/ibm_rans2p/comparison_files/T1_ibm_rans2p.h5']),
                      (os.path.join(proteus_install_path,'tests','cylinder2D','ibm_rans2p_3D','comparison_files'),
                       ['proteus/tests/cylinder2D/ibm_rans2p_3D/comparison_files/T1_ibm_3D_rans2p.h5']),
                      (os.path.join(proteus_install_path,'tests','cylinder2D','sbm_method','comparison_files'),
                       ['proteus/tests/cylinder2D/sbm_method/comparison_files/T1_sbm_rans3p.h5']),
                      (os.path.join(proteus_install_path,'tests','cylinder2D','sbm_3Dmesh','comparison_files'),
                       ['proteus/tests/cylinder2D/sbm_3Dmesh/comparison_files/T001_P1_sbm_3Dmesh.h5']),
                      (os.path.join(proteus_install_path,'tests','cylinder2D','conforming_rans2p','comparison_files'),
                       ['proteus/tests/cylinder2D/conforming_rans2p/comparison_files/T1_rans2p.h5']),
                      (os.path.join(proteus_install_path,'tests','HotStart_3P','comparison_files'),
                       ['proteus/tests/HotStart_3P/comparison_files/T01P1_hotstart.h5',
                        'proteus/tests/HotStart_3P/comparison_files/T01P2_hotstart.h5']),
                      (os.path.join(proteus_install_path,'tests','AddedMass'),
                       ['proteus/tests/AddedMass/petsc.options.superlu_dist']),
                      (os.path.join(proteus_install_path,'tests','MoveMeshMonitor'),
                       ['proteus/tests/MoveMeshMonitor/petsc.options.asm',
                        'proteus/tests/MoveMeshMonitor/nodesResult.csv']),
                       (os.path.join(proteus_install_path,'tests','wave_tests'),
                       ['proteus/tests/wave_tests/data_timeSeries.dat',
                        'proteus/tests/wave_tests/data_timeSeries.txt',
                        'proteus/tests/wave_tests/data_timeSeries_err1.csv',
                        'proteus/tests/wave_tests/data_timeSeries_err2.txt']),
         ],
          scripts = ['scripts/parun','scripts/gf2poly','scripts/gatherArchives.py','scripts/qtm','scripts/waves2xmf','scripts/povgen.py',
                     'scripts/velocity2xmf','scripts/run_script_garnet','scripts/run_script_diamond',
                     'scripts/run_script_lonestar','scripts/run_script_ranger','scripts/run_script_mpiexec','scripts/gatherTimes.py'],
          requires=['numpy']
    )

def setup_extensions_in_sequential():
    setup_given_extensions(EXTENSIONS_TO_BUILD)

def setup_extensions_in_parallel():
    import multiprocessing, logging
    logger = multiprocessing.log_to_stderr()
    logger.setLevel(logging.INFO)
    multiprocessing.log_to_stderr()
    pool = multiprocessing.Pool(processes=multiprocessing.cpu_count())
    EXTENSIONS=[[e] for e in EXTENSIONS_TO_BUILD]
    pool.imap(setup_given_extensions, EXTENSIONS)
    pool.close()
    pool.join()

if "build_ext" in sys.argv:
    setup_extensions_in_parallel()
else:
    setup_extensions_in_sequential()
